"""Rattr file analyser."""
from __future__ import annotations

import ast
import copy
from collections import deque
from typing import TYPE_CHECKING

import attrs

from rattr import error
from rattr.analyser.base import NodeVisitor
from rattr.analyser.cls import ClassAnalyser
from rattr.analyser.function import FunctionAnalyser
from rattr.analyser.types import ImportIrs
from rattr.analyser.util import (
    has_annotation,
    is_excluded_name,
    parse_rattr_results_from_annotation,
    read,
    timer,
)
from rattr.ast.util import (
    assignment_is_one_to_one,
    assignment_targets,
    fullname_of,
    has_lambda_in_rhs,
    has_namedtuple_declaration_in_rhs,
    walruses_in_rhs,
)
from rattr.config import Config
from rattr.config.state import enter_file
from rattr.extra import DictChanges
from rattr.models.context import Context, compile_root_context
from rattr.models.ir import FileIr
from rattr.models.symbol import Import
from rattr.module_locator.util import is_in_import_blacklist, is_in_pip, is_in_stdlib
from rattr.plugins import plugins

if TYPE_CHECKING:
    from rattr.models.symbol import Func


@attrs.mutable
class RattrStats:
    parse_time: float
    root_context_time: float
    assert_time: float
    analyse_imports_time: float
    analyse_file_time: float

    file_lines: int
    import_lines: int

    number_of_imports: int
    number_of_unique_imports: int


@attrs.mutable
class RattrImportStats:
    import_lines: int
    number_of_imports: int
    number_of_unique_imports: int


def parse_and_analyse_file() -> tuple[FileIr, ImportIrs, RattrStats]:
    """Parse and analyse the target file from the config."""
    config = Config()

    with enter_file(config.arguments.target):
        file_ir, import_irs, stats = __parse_and_analyse_file_impl()

    return file_ir, import_irs, stats


def __parse_and_analyse_file_impl() -> tuple[FileIr, ImportIrs, RattrStats]:
    """Parse and analyse the given file contents."""
    config = Config()

    with timer() as parse_timer, read(config.arguments.target) as (file_lines, source):
        ast_module = ast.parse(source)

    with timer() as root_context_timer:
        context = compile_root_context(ast_module).expand_starred_imports()

    with timer() as assert_timer:
        for assertor in plugins.assertors:
            assertor.assert_holds(ast_module, copy.deepcopy(context))

    with timer() as analyse_imports_timer:
        if config.arguments.follow_imports:
            imports = [s for s in context.symbol_table.symbols if isinstance(s, Import)]
            import_irs, import_stats = parse_and_analyse_imports(imports)
        else:
            import_irs, import_stats = {}, RattrImportStats(0, 0, 0)

    with timer() as analyse_file_timer:
        file_ir = FileAnalyser(ast_module, context).analyse()

    stats = RattrStats(
        parse_time=parse_timer.time,
        root_context_time=root_context_timer.time,
        assert_time=assert_timer.time,
        analyse_imports_time=analyse_imports_timer.time,
        analyse_file_time=analyse_file_timer.time,
        file_lines=file_lines,
        import_lines=import_stats.import_lines,
        number_of_imports=import_stats.number_of_imports,
        number_of_unique_imports=import_stats.number_of_unique_imports,
    )
    return file_ir, import_irs, stats


def parse_and_analyse_imports(
    imports: list[Import],
) -> tuple[ImportIrs, RattrImportStats]:
    """Return the mapping from file name to IR for each import.

    Imports are a directed cyclic graph, however, previously analysed files can
    just be ignored (analysing is deterministic and context-free). Thus, the
    graph of imports becomes a DAG which we BFS.
    """
    config = Config()
    queue = deque(imports)

    import_irs: ImportIrs = {}
    import_stats = RattrImportStats(
        import_lines=0,
        number_of_imports=0,
        number_of_unique_imports=0,
    )

    seen_module_origins: set[str] = set()

    while queue:
        import_ = queue.popleft()
        import_stats.number_of_imports += 1

        name = import_.module_name
        spec = import_.module_spec

        if name is None:
            error.error(f"unable to resolve import {import_.qualified_name!r}")
            continue

        if spec is None:
            error.error(f"unable to resolve module spec for {name!r}")
            continue

        if spec.origin is None:
            # HACK Can't use isinstance
            # TODO Resolve BuiltinImporter modules
            if "BuiltinImporter" in str(spec.loader):
                error.error(f"unable to resolve builtin module {name!r}", badness=0)
            else:
                error.error(f"unable to resolve import {import_.qualified_name!r}")
            continue

        if spec.origin in seen_module_origins:
            continue

        if is_in_import_blacklist(name):
            continue

        if not config.arguments.follow_pip_imports and is_in_pip(name):
            continue

        if not config.arguments.follow_stdlib_imports and is_in_stdlib(name):
            continue

        with read(spec.origin) as (import_file_lines, import_file_source):
            import_ast = ast.parse(import_file_source)

        with enter_file(spec.origin):
            import_context = compile_root_context(import_ast).expand_starred_imports()
            import_ir = FileAnalyser(import_ast, import_context).analyse()

        import_irs[name] = import_ir

        imports_in_the_current_import = [
            symbol
            for symbol in import_context.symbol_table.symbols
            if isinstance(symbol, Import)
        ]
        for next_import in imports_in_the_current_import:
            queue.append(next_import)

        import_stats.import_lines += import_file_lines

        seen_module_origins.add(spec.origin)

    import_stats.number_of_unique_imports = len(seen_module_origins)
    return import_irs, import_stats


class FileAnalyser(NodeVisitor):
    """Walk a file's AST and analyse the contained functions and classes."""

    def __init__(self, _ast: ast.Module, context: Context) -> None:
        """Set configuration and initialise results."""
        self._ast = _ast
        self.context = context
        self.file_ir = FileIr(context=context)

    def analyse(self) -> FileIr:
        """Entry point of FileAnalyser, return the results of analysis."""
        self.visit(self._ast)

        return self.file_ir

    def visit_AnyFunctionDef(
        self,
        node: ast.FunctionDef | ast.AsyncFunctionDef,
    ) -> None:
        if has_annotation("rattr_ignore", node):
            return

        if is_excluded_name(node.name):
            return

        try:
            fn = self.context.get_func_or_error(node.name)
        except KeyError as exc:
            return error.error(str(exc.args[0]), culprit=node)

        if has_annotation("rattr_results", node):
            return self.visit_function_with_rattr_results_annotation(node, fn)

        if plugins.has_analyser(fn, modulename=self.context.modulename):
            return self.visit_function_with_custom_analyser(node, fn)

        self.file_ir[fn] = FunctionAnalyser(node, self.context).analyse()

    def visit_function_with_rattr_results_annotation(
        self,
        node: ast.FunctionDef | ast.AsyncFunctionDef,
        fn: Func,
    ) -> None:
        self.file_ir[fn] = parse_rattr_results_from_annotation(
            node,
            context=self.context,
        )

    def visit_function_with_custom_analyser(
        self,
        node: ast.FunctionDef | ast.AsyncFunctionDef,
        fn: Func,
    ) -> None:
        custom_analyser = plugins.get_analyser(fn, modulename=self.context.modulename)
        self.file_ir[fn] = custom_analyser.on_def(fn.id, node, self.context)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self.visit_AnyFunctionDef(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self.visit_AnyFunctionDef(node)

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        if has_annotation("rattr_ignore", node):
            return

        if is_excluded_name(node.name):
            return

        class_ir = ClassAnalyser(node, self.context).analyse()

        for foc, foc_ir in class_ir.items():
            self.file_ir[foc] = foc_ir

    # ----------------------------------------------------------------------- #
    # Lambdas
    # ----------------------------------------------------------------------- #

    def visit_LambdaAssign(
        self,
        node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
    ) -> None:
        if not assignment_is_one_to_one(node):
            return error.fatal("lambda assignment must be one-to-one", node)

        targets = assignment_targets(node)
        name = fullname_of(targets[0])

        try:
            fn = self.context.get_func_or_error(name)
        except KeyError as exc:
            return error.error(str(exc.args[0]), culprit=node)

        self.file_ir[fn] = FunctionAnalyser(node.value, self.context).analyse()

    def visit_NamedTupleAssign(
        self,
        node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
    ) -> None:
        if not assignment_is_one_to_one(node):
            return error.fatal("namedtuple assignment must be one-to-one", node)

        name = fullname_of(assignment_targets(node)[0])

        try:
            cls = self.context.get_class_or_error(name)
        except KeyError as exc:
            return error.error(str(exc.args[0]), culprit=node)

        self.file_ir[cls] = {
            "gets": set(),
            "sets": set(),
            "dels": set(),
            "calls": set(),
        }

    def visit_AnyAssign(
        self,
        node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
    ) -> None:
        if has_lambda_in_rhs(node):
            self.visit_LambdaAssign(node)

        if has_namedtuple_declaration_in_rhs(node):
            self.visit_NamedTupleAssign(node)

        # Walrus may obscure a lambda, so peek in and visit the nice walruses
        for walrus in walruses_in_rhs(node):
            with DictChanges(self.file_ir) as diff:
                self.visit_AnyAssign(walrus)

            # Handle `outer_lhs = (inner_lhs := lambda: ...)`
            if has_lambda_in_rhs(walrus):
                if len(diff.added) == 1 and node.value == walrus:
                    inner_rhs = list(diff.added)[0]
                    outer_lhs = attrs.evolve(
                        inner_rhs,
                        name=fullname_of(assignment_targets(node)[0]),
                    )
                    self.file_ir[outer_lhs] = self.file_ir[inner_rhs]
                elif len(diff.added) > 1:
                    raise NotImplementedError("Multiple deeply nested walruses")

    def visit_Assign(self, node: ast.Assign) -> None:
        self.visit_AnyAssign(node)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        self.visit_AnyAssign(node)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        self.visit_AnyAssign(node)

    def visit_NamedExpr(self, node: ast.NamedExpr) -> None:
        self.visit_AnyAssign(node)

    def visit_Lambda(self, node: ast.Lambda) -> None:
        # NOTE Only reached when the lambda is an anonymous function
        return error.fatal("module level lambdas unsupported", node)



from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

from rattr.ast.types import (  # noqa: F401
    Identifier,
)
from rattr.models.ir import FileIr, FunctionIr
from rattr.models.symbol import UserDefinedCallableSymbol
from rattr.versioning.typing import TypeAlias

if TYPE_CHECKING:
    from rattr.versioning.typing import TypeAlias


ClassIr: TypeAlias = dict[UserDefinedCallableSymbol, FunctionIr]

_ModuleName: TypeAlias = str
ImportIrs: TypeAlias = dict[_ModuleName, FileIr]


TargetName: TypeAlias = Identifier
PositionalArgumentName: TypeAlias = Identifier
KeywordArgumentName: TypeAlias = Identifier
LocalIdentifier: TypeAlias = Identifier


class RattrResults(TypedDict):
    gets: set[Identifier]
    sets: set[Identifier]
    dels: set[Identifier]
    calls: list[
        tuple[
            TargetName,
            tuple[
                list[PositionalArgumentName],
                dict[KeywordArgumentName, LocalIdentifier],
            ],
        ]
    ]



from __future__ import annotations

# isort: off
from ._types import (
    Arguments,
    Config,
    FollowImports,
    Output,
    FormatPath,
    ShowWarnings,
    State,
)

__all__ = [
    "Arguments",
    "Config",
    "FollowImports",
    "Output",
    "FormatPath",
    "ShowWarnings",
    "State",
]



from __future__ import annotations

import ast
from itertools import product
from typing import TYPE_CHECKING, Final

import pytest

from rattr.ast.types import AstLiterals
from rattr.models.context._util import (
    is_call_to_call_result,
    is_call_to_literal,
    is_call_to_member_of_module_import,
    is_call_to_method,
    is_call_to_method_on_primitive_from_call,
    is_call_to_method_on_py_type,
    is_call_to_subscript_item,
    is_direct_call_to_method_on_constant,
    is_direct_call_to_method_on_literal,
)
from rattr.models.symbol._symbol import Symbol

if TYPE_CHECKING:
    from rattr.models.symbol import Import

symbol_fixtures: Final = (
    "simple_name",
    "simple_builtin",
    "simple_import",
    "simple_func",
    "simple_class",
    "simple_call",
)


class TestIsCallToLiteral:
    def test_literal(self, constant):
        assert is_call_to_literal(constant)

    @pytest.mark.parametrize("fn", ["some_function"])
    def test_non_literal(self, fn):
        assert not is_call_to_literal(fn)
        assert not is_call_to_literal(f"{fn}()")


class TestIsCallToSubscriptItem:
    def test_literal(self, constant):
        assert not is_call_to_subscript_item(constant)

    @pytest.mark.parametrize("fn", ["some_function"])
    def test_non_literal(self, fn):
        assert not is_call_to_subscript_item(fn)
        assert not is_call_to_subscript_item(f"{fn}()")

    def test_constant_subscript_element(self, constant):
        assert is_call_to_subscript_item(f"{constant}[]")
        assert is_call_to_subscript_item(f"{constant}[]()")

    def test_constant_subscript_element_method(self, constant):
        assert is_call_to_subscript_item(f"{constant}[].method")
        assert is_call_to_subscript_item(f"{constant}[].method()")

    def test_literal_subscript_element(self, literal):
        assert is_call_to_subscript_item(f"{literal(ast.List)}[]")
        assert is_call_to_subscript_item(f"{literal(ast.List)}[]()")

    def test_literal_subscript_element_method(self, literal):
        assert is_call_to_subscript_item(f"{literal(ast.List)}[].method")
        assert is_call_to_subscript_item(f"{literal(ast.List)}[].method()")

    @pytest.mark.parametrize("fn", ["my_variable[]"])
    def test_to_variable_subscript_element(self, fn):
        assert is_call_to_subscript_item(f"{fn}")
        assert is_call_to_subscript_item(f"{fn}()")

    @pytest.mark.parametrize("fn", ["my_variable[].method"])
    def test_to_variable_subscript_element_method(self, fn):
        assert is_call_to_subscript_item(f"{fn}")
        assert is_call_to_subscript_item(f"{fn}()")


class TestIsCallToMethod:
    def test_null_call_positive(self):
        assert is_call_to_method(None, "nothing", None, "nothing.method()")

    def test_null_call_negative(self):
        assert not is_call_to_method(None, "nothing", None, "nothing")

    @pytest.mark.parametrize(
        "symbol_fixture",
        [s for s in symbol_fixtures if s != "simple_import"],
    )
    def test_null_call_with_valid_lhs(
        self,
        request: pytest.FixtureRequest,
        symbol_fixture: str,
    ):
        # I.e. target was `my_variable.method()`
        # S.t. symbol = context.get("my_variable.method()") i.e. None
        # &    lsh_symbol = context.get("my_variable") i.e. from fixture
        # Excludes the case of an import, as that is not a method call!
        lhs_symbol: Symbol = request.getfixturevalue(symbol_fixture)

        assert is_call_to_method(
            None,
            f"{lhs_symbol.name}.method()",
            lhs_symbol,
            lhs_symbol.name,
        )

    def test_null_call_with_import_lhs(self, simple_import: Import):
        assert not is_call_to_method(None, "jimmy", simple_import, simple_import.name)

    @pytest.mark.parametrize(
        "symbol_fixture,lhs_symbol_fixture",
        [pair for pair in product(symbol_fixtures, symbol_fixtures)],
    )
    def test_valid_target(
        self,
        request: pytest.FixtureRequest,
        symbol_fixture: str,
        lhs_symbol_fixture: str,
    ):
        symbol: Symbol = request.getfixturevalue(symbol_fixture)
        lhs_symbol: Symbol = request.getfixturevalue(lhs_symbol_fixture)
        assert not is_call_to_method(symbol, symbol.name, lhs_symbol, lhs_symbol.name)


class TestIsCallToMemberOfModuleImport:
    @pytest.mark.parametrize("symbol_fixture", symbol_fixtures)
    def test_non_null_target(
        self,
        request: pytest.FixtureRequest,
        symbol_fixture: str,
    ):
        symbol: Symbol = request.getfixturevalue(symbol_fixture)

        assert not is_call_to_member_of_module_import("pi", symbol)
        assert not is_call_to_member_of_module_import("math.pi", symbol)

    def test_null_target(self):
        assert not is_call_to_member_of_module_import("pi", None)
        assert is_call_to_member_of_module_import("math.pi", None)


class TestIsDirectCallToMethodOnConstant:
    def test_call_to_method_on_constant(self, constant):
        assert not is_direct_call_to_method_on_constant(constant)
        assert is_direct_call_to_method_on_constant(f"{constant}.method()")

    @pytest.mark.parametrize("literal_type", AstLiterals)
    def test_call_to_method_on_literal(self, literal, literal_type):
        as_str = literal(literal_type)
        assert not is_direct_call_to_method_on_constant(as_str)
        assert not is_direct_call_to_method_on_constant(f"{as_str}.method()")

    @pytest.mark.parametrize("name", ["@Constant.pop()"])
    def test_direct_call_to_method_on_constant(self, name):
        assert is_direct_call_to_method_on_constant(name)

    @pytest.mark.parametrize("name", ["@Constant[].my_method()"])
    def test_indirect_call_to_method_on_constant(self, name):
        assert not is_direct_call_to_method_on_constant(name)

    @pytest.mark.parametrize("name", ["Constant.but_not_really()"])
    def test_is_not_call_to_method_on_constant(self, name):
        assert not is_direct_call_to_method_on_constant(name)


class TestIsDirectCallToMethodOnLiteral:
    def test_call_to_method_on_constant(self, constant):
        assert not is_direct_call_to_method_on_literal(constant)
        assert not is_direct_call_to_method_on_literal(f"{constant}.method()")

    @pytest.mark.parametrize("literal_type", AstLiterals)
    def test_direct_call_to_method_on_literal(self, literal, literal_type):
        as_str = literal(literal_type)
        assert not is_direct_call_to_method_on_literal(as_str)
        assert is_direct_call_to_method_on_literal(f"{as_str}.method()")

    @pytest.mark.parametrize("literal_type", AstLiterals)
    def test_indirect_call_to_method_on_literal(self, literal, literal_type):
        as_str = literal(literal_type)
        assert not is_direct_call_to_method_on_literal(f"{as_str}[].my_method()")

    @pytest.mark.parametrize("literal_type", AstLiterals)
    def test_is_not_call_to_method_on_literal(self, literal_type):
        as_str = literal_type.__name__
        assert not is_direct_call_to_method_on_literal(f"{as_str}.but_not_really()")


class TestIsCallToMethodOnPrimitiveFromCall:
    @pytest.mark.parametrize(
        "expr",
        [
            "int().to_bytes()",
            "str().split()",
            "int.to_bytes()",
            "str.split()",
        ],
    )
    def test_positive_case(self, expr):
        assert is_call_to_method_on_primitive_from_call(expr)

    @pytest.mark.parametrize(
        "expr",
        ["my_int.to_bytes()", "my_int().to_bytes()"],
    )
    def test_negative_case(self, expr):
        assert not is_call_to_method_on_primitive_from_call(expr)


class TestIsCallToMethodOnPyType:
    @pytest.mark.parametrize(
        "expr",
        [
            "@Constant.removeprefix()",
            "@List.append()",
            "float().from_hex()",
        ],
    )
    def test_positive_case(self, expr):
        assert is_call_to_method_on_py_type(expr)

    @pytest.mark.parametrize(
        "expr",
        [
            "@NotConstant.removeprefix()",
            "@NotList.append()",
            "my_float().from_hex()",
        ],
    )
    def test_negative_case(self, expr):
        assert not is_call_to_method_on_py_type(expr)


class TestIsCallToCallResult:
    @pytest.mark.parametrize("expr", ["my_var()()"])
    def test_positive_case(self, expr):
        assert is_call_to_call_result(ast.parse(expr).body[0].value)

    @pytest.mark.parametrize("expr", ["my_var().method()", "my_var[0]()"])
    def test_negative_case(self, expr):
        assert not is_call_to_call_result(ast.parse(expr).body[0].value)



from __future__ import annotations

from rattr.analyser.annotations import rattr_ignore, rattr_results


def test_rattr_ignore_is_identity():
    @rattr_ignore()
    def my_test_fn(a: str) -> bool:
        return a == "hello"

    assert not my_test_fn("hi")
    assert my_test_fn("hello")


def test_rattr_ignore_without_call_brackets_is_identity():
    @rattr_ignore
    def my_test_fn(a: str) -> bool:
        return a == "hello"

    assert not my_test_fn("hi")
    assert my_test_fn("hello")


def test_rattr_results_is_identity():
    @rattr_results(gets={"a", "b"})
    def my_test_fn(a: str) -> bool:
        return a == "hello"

    assert not my_test_fn("hi")
    assert my_test_fn("hello")



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from rattr.analyser.util import read

if TYPE_CHECKING:
    from typing import Callable


here = Path(__file__).resolve().parent


@pytest.fixture
def get_test_file() -> Callable[[str], Path]:
    def __file_getter(name: str):
        the_file = here / "data" / name
        assert the_file.exists(), f"{str(the_file)} does not exist"
        return the_file

    return __file_getter


@pytest.fixture
def get_non_existent_file() -> Callable[[str], Path]:
    def __file_getter(name: str):
        the_file = here / "data" / name
        assert not the_file.exists(), f"{str(the_file)} exists"
        return the_file

    return __file_getter


class TestRead:
    @pytest.mark.parametrize(
        "filename",
        ["the_empty_file.py", "only_blank_lines.py", "a_populated_file.py"],
    )
    def test_file(self, get_test_file, filename):
        the_file: Path = get_test_file(filename)

        with read(the_file) as (number_of_lines, content):
            assert content == the_file.read_text()
            assert number_of_lines == len(the_file.read_text().splitlines()) + 1

    @pytest.mark.parametrize("filename", ["foo.py"])
    def test_non_existent_file(self, get_non_existent_file, filename):
        with pytest.raises(FileNotFoundError):
            with read(get_non_existent_file(filename)) as _:
                ...



from __future__ import annotations

import importlib.util
from typing import TYPE_CHECKING

import pytest

from rattr.module_locator.models import ModuleSpec
from rattr.module_locator.util import (
    derive_module_names_left,
    derive_module_names_right,
    find_module_name_and_spec,
    find_module_spec_fast,
    format_origin_for_os,
    is_in_import_blacklist,
    is_in_pip,
    is_in_stdlib,
)

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Final

    from tests.shared import ArgumentsFn


known_targets_in_stdlib_modules: Final = (
    "os.path",
    "os.path.join",
    "math.sin",
    "math.pi",
)

known_pip_modules: Final = (
    "isort",
    "pytest",
)
known_pip_targets: Final = (
    "isort.place.place_module",
    "pytest.mark.parameterize",
)
known_pip_modules_absent_in_venv: Final = (
    "numpy",
    "pandas",
)

known_rattr_modules: Final = (
    "rattr",
    "rattr.ast.util",
    "rattr.models.context",
)
known_rattr_targets: Final = (
    "rattr.ast.util.basename_of",
    "rattr.models._context.Context",
)


@pytest.mark.parametrize(
    "target,expected",
    testcases := [
        # the empty string
        ("", [""]),
        # general
        ("non_dotted_name", ["non_dotted_name"]),
        ("a.b.c.d", ["a.b.c.d", "b.c.d", "c.d", "d"]),
        # stdlib
        ("os.path.join", ["os.path.join", "path.join", "join"]),
        ("pathlib.Path", ["pathlib.Path", "Path"]),
        # internal
        ("rattr.util.is_name", ["rattr.util.is_name", "util.is_name", "is_name"]),
    ],
    ids=[t[0] for t in testcases],
)
def test_derive_module_names_left(target: str, expected: list[str]):
    assert derive_module_names_left(target) == expected


@pytest.mark.parametrize(
    "target,expected",
    testcases := [
        # the empty string
        ("", [""]),
        # general
        ("non_dotted_name", ["non_dotted_name"]),
        ("a.b.c.d", ["a.b.c.d", "a.b.c", "a.b", "a"]),
        # stdlib
        ("os.path.join", ["os.path.join", "os.path", "os"]),
        ("pathlib.Path", ["pathlib.Path", "pathlib"]),
        # internal
        ("rattr.util.is_name", ["rattr.util.is_name", "rattr.util", "rattr"]),
    ],
    ids=[t[0] for t in testcases],
)
def test_derive_module_names_right(target: str, expected: list[str]):
    assert derive_module_names_right(target) == expected


@pytest.mark.parametrize(
    "target",
    [
        # the empty string
        "",
        # non-existant module
        "i.am.not.a.module",
    ],
)
def test_find_module_name_and_spec_non_existant(target: str):
    assert find_module_name_and_spec(target) == (None, None)


@pytest.mark.parametrize(
    "target,module_name",
    [
        # stdlib
        ("math", "math"),
        ("math.pi", "math"),
        ("os", "os"),
        ("os.path", "os.path"),
        # internal
        ("rattr.analyser", "rattr.analyser"),
        ("rattr.analyser.util.get_module_spec", "rattr.analyser.util"),
    ],
)
def test_find_module_name_and_spec_real_world(target: str, module_name: str):
    # NOTE
    # Use `find_spec` not `find_module_spec_fast` here as we are testing the correctness
    # of the latter's impl.
    stdlib_spec = importlib.util.find_spec(module_name)
    rattr_spec = (
        ModuleSpec(
            name=stdlib_spec.name,
            origin=format_origin_for_os(stdlib_spec.origin),
        )
        if stdlib_spec is not None
        else None
    )
    assert find_module_name_and_spec(target) == (module_name, rattr_spec)


@pytest.mark.parametrize(
    "modulename",
    testcases := [
        # non-existant
        "",
        "does_not_exist",
        "does.not.exist",
        # stdlib
        "math",
        "math.pi",
        "os.path",
        # local (rattr)
        "rattr.does.not.exist",
        "rattr.ast",
        "rattr.ast.util",
        "rattr.analyser.annotations.rattr_ignore",
        # pip installed
        "isort",
        "isort.place",
        "attrs",
    ],
    ids=testcases,
)
def test_find_module_spec_fast(modulename: str):
    try:
        stdlib_spec = importlib.util.find_spec(modulename)
    except (AttributeError, ModuleNotFoundError, ValueError):
        stdlib_spec = None

    expected = (
        ModuleSpec(
            name=stdlib_spec.name,
            origin=format_origin_for_os(stdlib_spec.origin),
        )
        if stdlib_spec is not None
        else None
    )

    assert find_module_spec_fast(modulename) == expected


@pytest.mark.linux
@pytest.mark.parametrize(
    "modulename, expected",
    testcases := [
        # stdlib
        ("sys", importlib.util.find_spec("sys").origin),
        ("math", importlib.util.find_spec("math").origin),
        ("os.path", importlib.util.find_spec("os.path").origin),
        # local (rattr)
        ("rattr.ast", importlib.util.find_spec("rattr.ast").origin),
        ("rattr.ast.util", importlib.util.find_spec("rattr.ast.util").origin),
        # pip installed
        (
            "isort",
            importlib.util.find_spec("isort").origin,
        ),
        (
            "isort.place",
            importlib.util.find_spec("isort.place").origin,
        ),
        (
            "attrs",
            importlib.util.find_spec("attrs").origin,
        ),
    ],
    ids=[t[0] for t in testcases],
)
def test_find_module_spec_fast_origins(modulename: str, expected: str):
    assert find_module_spec_fast(modulename).origin == expected


@pytest.mark.parametrize(
    "modulename, expected",
    testcases := [
        ("", False),
        ("anything.dotted.anything", False),
        *[(stdlib_target, True) for stdlib_target in known_targets_in_stdlib_modules],
        ("math.this.is.not.in.the.stdlib", True),
        *[
            (pip_target, False)
            for pip_target in known_pip_modules
            + known_pip_targets
            + known_pip_modules_absent_in_venv
        ],
        *[
            (rattr_target, False)
            for rattr_target in known_rattr_modules + known_rattr_targets
        ],
    ],
    ids=[t[0] for t in testcases],
)
def test_is_in_stdlib(modulename: str, expected: bool):
    assert is_in_stdlib(modulename) == expected


def test_is_in_stdlib_known_stdlib_modules(stdlib_modules: Iterable[str]):
    for modulename in stdlib_modules:
        assert is_in_stdlib(modulename)


@pytest.mark.parametrize(
    "modulename, expected",
    testcases := [
        ("", False),
        ("anything.dotted.anything", False),
        *[(stdlib_target, False) for stdlib_target in known_targets_in_stdlib_modules],
        ("math.this.is.not.in.the.stdlib", False),
        *[(pip_target, True) for pip_target in known_pip_modules + known_pip_targets],
        *[(pip_target, False) for pip_target in known_pip_modules_absent_in_venv],
        *[
            (rattr_target, False)
            for rattr_target in known_rattr_modules + known_rattr_targets
        ],
    ],
    ids=[t[0] for t in testcases],
)
def test_is_in_pip(modulename: str, expected: bool):
    assert is_in_pip(modulename) == expected


def test_is_in_pip_known_stdlib_modules(stdlib_modules: Iterable[str]):
    for modulename in stdlib_modules:
        assert not is_in_pip(modulename)


@pytest.mark.parametrize(
    "modulename, expected",
    testcases := [
        ("", True),
        ("anything.dotted.anything", False),
        *[(stdlib_target, False) for stdlib_target in known_targets_in_stdlib_modules],
        ("math.this.is.not.in.the.stdlib", False),
        *[
            (pip_target, False)
            for pip_target in known_pip_modules
            + known_pip_targets
            + known_pip_modules_absent_in_venv
        ],
        *[
            (rattr_target, True)
            for rattr_target in known_rattr_modules + known_rattr_targets
        ],
    ],
    ids=[t[0] for t in testcases],
)
def test_is_in_import_blacklist(modulename: str, expected: bool):
    assert is_in_import_blacklist(modulename) == expected


def test_is_in_import_blacklist_known_stdlib_modules(stdlib_modules: Iterable[str]):
    for modulename in stdlib_modules:
        assert not is_in_import_blacklist(modulename)


@pytest.mark.parametrize(
    "banned,unbanned",
    [
        (
            {"beelzebub", "abaddon"},
            {"michael", "raphael", "gabriel"},
        )
    ],
)
def test_is_in_import_blacklist_from_cli_argument(
    arguments: ArgumentsFn,
    banned: set[str],
    unbanned: set[str],
):
    is_in_import_blacklist.cache_clear()

    # Nothing is banned
    with arguments(_excluded_imports=set()):
        for banned_module in sorted(banned):
            assert not is_in_import_blacklist(banned_module)
        for unbanned_module in sorted(unbanned):
            assert not is_in_import_blacklist(unbanned_module)

    is_in_import_blacklist.cache_clear()

    # Only bad things are banned
    with arguments(_excluded_imports=banned):
        for banned_module in sorted(banned):
            assert is_in_import_blacklist(banned_module)
        for unbanned_module in sorted(unbanned):
            assert not is_in_import_blacklist(unbanned_module)



from __future__ import annotations

import ast
from typing import TYPE_CHECKING

import attrs

from rattr import error
from rattr.ast.util import (
    assignment_is_one_to_one,
    assignment_targets,
    fullname_of,
    has_lambda_in_rhs,
    has_namedtuple_declaration_in_rhs,
    is_relative_import,
    is_starred_import,
    namedtuple_init_signature_from_declaration,
    walruses_in_rhs,
)
from rattr.codegen import gen_import_from_stmt
from rattr.config import Config
from rattr.extra import DictChanges
from rattr.models.context._context import Context
from rattr.models.context._symbol_table import SymbolTable
from rattr.models.symbol._symbols import (
    PYTHON_BUILTINS,
    Builtin,
    CallInterface,
    Class,
    Func,
    Import,
    Name,
)
from rattr.module_locator.util import (
    derive_absolute_module_name,
    derive_module_name_from_path,
    find_module_name_and_spec,
    is_in_import_blacklist,
)

if TYPE_CHECKING:
    from typing import Final

    from rattr.ast.types import Identifier
    from rattr.module_locator.util import ModuleName


# TODO
#   It would be nice to derive these automatically as with the builtins.
#   However, this should always be the union of the appropriate attrs for all supported
#   Python versions as we have no way of knowing which version of Python the targeted
#   source file is using.
MODULE_LEVEL_DUNDER_ATTRS: Final = (
    "__annotations__",
    "__builtins__",
    "__cached__",
    "__doc__",
    "__file__",
    "__loader__",
    "__name__",
    "__package__",
    "__spec__",
)
"""The dunder attrs provided by the interpreter at module level, such as `__file``."""


def compile_root_context(module: ast.Module) -> Context:
    if not isinstance(module, ast.Module):
        raise TypeError("The root context must be derived from a module")

    root = Context(parent=None, symbol_table=SymbolTable())

    # Add the Python default names/builtins with dummy tokens (for a consistent symbol
    # interface).
    root.add([__module_level_name(name) for name in MODULE_LEVEL_DUNDER_ATTRS])
    root.add([__module_level_builtin(name) for name in PYTHON_BUILTINS])

    # Populate the context with the top-level declarations
    _root_context_builder = RootContextBuilder(context=root)
    _root_context_builder.register_stmts(*module.body)

    return root


class RootContextBuilder:
    def __init__(self, context: Context) -> None:
        self.context: Context = context
        super().__init__()

    def register(self, node: ast.AST) -> None:
        if isinstance(node, ast.Module):
            return TypeError("use register_stmts(module.body) for modules")

        # NOTE
        # We do this and implement "visit_If", etc rather than inheriting from
        # `ast.NodeVisitor` as we only want to visit the top-level (i.e. not statements
        # in function or class definitions) statements not every
        # node at every depth.
        visit_method_name = f"visit_{node.__class__.__name__}"
        visit = getattr(self, visit_method_name, None)

        if visit is None:
            return

        return visit(node)

    def register_stmts(self, *stmts: ast.stmt) -> None:
        for stmt in stmts:
            self.register(stmt)

    # ================================================================================ #
    # Registration methods
    # ================================================================================ #

    def visit_Import(self, node: ast.Import) -> None:
        """Register an import.

        >>> import math
        Import(name="math", qualified_name="math")

        >>> import math as m
        Import(name="m", qualified_name="math")

        >>> import math, more_math
        Import(name="math", qualified_name="math")
        Import(name="more_math", qualified_name="more_math")
        """
        if len(node.names) > 1:
            error.info("do not import multiple modules on one line", culprit=node)

        self.context.add(
            make_import_symbol(
                name=alias.asname or alias.name,
                qualified_name=alias.name,
                module_name=alias.name,
                token=node,
            )
            for alias in node.names
        )

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        """Register an import-from.

        >>> # Regular (named) import
        >>> from math import pi
        Import(name="pi", qualified_name="math.pi")

        >>> # Regular (named) import with alias
        >>> from math import pi as pie
        Import(name="pie", qualified_name="math.pi")

        >>> # Starred-import
        >>> from math import *
        Import(name="*", qualified_name="math")

        >>> # Relative import
        >>> from .relative.module import my_function

        >>> # Relative starred import
        >>> from .relative.module import *

        >>> # Relative starred import from local dir
        >>> from . import module
        """
        # Warn when using starred imports outside of __init__.py
        # There is little good about starred imports (and no benefit over
        # `import numpy as np`, for example), and many issues such as clobbering, etc.
        if is_starred_import(node) and not self.context.is_init_file:
            error_starred_import_outside_init(node, node.module or node.names[0].name)

        # Dispatch to specific import-type handler
        if is_relative_import(node) and is_starred_import(node):
            return self.visit_starred_relative_import(node)
        elif is_relative_import(node):
            return self.visit_relative_import(node)
        elif is_starred_import(node):
            return self.visit_starred_import(node)
        else:
            return self.visit_named_import(node)

    def visit_starred_relative_import(self, node: ast.ImportFrom) -> None:
        config = Config()
        base = derive_module_name_from_path(config.state.current_file)

        if base is None:
            raise ValueError  # only when the file can't be a module, so never here

        module_name = derive_absolute_module_name(base, node.module, node.level)
        (confirmed_module_name, spec) = find_module_name_and_spec(module_name)

        if spec is None:
            error.error("unable to resolve relative starred import", culprit=node)

        if (confirmed_module_name, spec) != (None, None):
            assert module_name == confirmed_module_name  # should always pass

        self.context.add(
            make_import_symbol(
                name="*",
                qualified_name=module_name,
                module_name=module_name,
                token=node,
            )
        )

    def visit_relative_import(self, node: ast.ImportFrom) -> None:
        config = Config()
        base = derive_module_name_from_path(config.state.current_file)

        if base is None:
            raise ValueError  # only when the file can't be a module, so never here

        module_name = derive_absolute_module_name(base, node.module, node.level)
        (confirmed_module_name, spec) = find_module_name_and_spec(module_name)

        if spec is None:
            error.error("unable to resolve relative import", culprit=node)

        if (confirmed_module_name, spec) != (None, None):
            assert module_name == confirmed_module_name  # should always pass

        self.context.add(
            make_import_symbol(
                name=target.asname or target.name,
                qualified_name=f"{module_name}.{target.name}",
                module_name=module_name,
                token=node,
            )
            for target in node.names
        )

    def visit_starred_import(self, node: ast.ImportFrom) -> None:
        if node.module is None:
            error.fatal("node has no module", culprit=node)

        self.context.add(
            make_import_symbol(
                name="*",
                qualified_name=node.module,
                module_name=node.module,
                token=node,
            )
        )

    def visit_named_import(self, node: ast.ImportFrom) -> None:
        if node.module is None:
            error.fatal("node has no module", culprit=node)

        self.context.add(
            make_import_symbol(
                name=target.asname or target.name,
                qualified_name=f"{node.module}.{target.name}",
                module_name=node.module,
                token=node,
            )
            for target in node.names
        )

    def visit_Assign(self, node: ast.Assign) -> None:
        self.visit_assignment(node)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        self.visit_assignment(node)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        self.visit_assignment(node)

    def visit_NamedExpr(self, node: ast.NamedExpr) -> None:
        self.visit_assignment(node)

    def visit_assignment(
        self,
        node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
    ) -> None:
        targets = assignment_targets(node)

        if has_lambda_in_rhs(node):
            if not assignment_is_one_to_one(node):
                return error.error("lambda assignment must be one-to-one", culprit=node)

            name = fullname_of(targets[0])
            interface = CallInterface.from_fn_def(node.value)

            self.context.add(Func(name=name, token=node, interface=interface))

        elif has_namedtuple_declaration_in_rhs(node):
            if not assignment_is_one_to_one(node):
                return error.error(
                    "namedtuple assignment must be one-to-one",
                    culprit=node,
                )

            name = fullname_of(targets[0])

            try:
                arguments = namedtuple_init_signature_from_declaration(node)
            except ValueError as exc:
                return error.error(str(exc.args[0]), culprit=node)

            interface = CallInterface(args=arguments)

            self.context.add(Class(name, token=node, interface=interface))

        else:
            for walrus in walruses_in_rhs(node):
                with DictChanges(
                    self.context.symbol_table,
                    iter_items=lambda t: t.values(),
                ) as diff:
                    self.visit_NamedExpr(walrus)

                # Handle `outer_lhs = (inner_lhs := lambda: ...)`
                if has_lambda_in_rhs(walrus):
                    if len(diff.added) == 1 and node.value == walrus:
                        inner_rhs = list(diff.added)[0]
                        outer_lhs = attrs.evolve(
                            inner_rhs,
                            name=fullname_of(assignment_targets(node)[0]),
                        )
                        self.context.add(outer_lhs)
                    elif len(diff.added) > 1:
                        error.error("multiple deeply nested walrus assignments")
                        continue

            for target in targets:
                self.context.add_identifiers_to_context(target)

    def visit_Delete(self, node: ast.Delete) -> None:
        for target in node.targets:
            self.context.remove_identifiers_from_context(target)

        error.warning(
            "avoid using 'del' at the module level as the target will be undefined so "
            "far as rattr is concerned",
            culprit=node,
        )

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self.context.add(Func.from_fn_def(node))

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self.context.add(Func.from_fn_def(node))

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        self.context.add(Class.from_class_def(node))

    def visit_If(self, node: ast.If) -> None:
        self.register_stmts(*node.body, *node.orelse)

    def visit_For(self, node: ast.For) -> None:
        self.register_stmts(*node.body, *node.orelse)

    def visit_AsyncFor(self, node: ast.AsyncFor) -> None:
        self.register_stmts(*node.body, *node.orelse)

    def visit_While(self, node: ast.While) -> None:
        self.register_stmts(*node.body, *node.orelse)

    def visit_Try(self, node: ast.Try) -> None:
        self.register_stmts(
            *node.body,
            *node.orelse,
            *node.finalbody,
            *(stmt for handler in node.handlers for stmt in handler.body),
        )

    def visit_With(self, node: ast.With) -> None:
        self.register_stmts(*node.body)

    def visit_AsyncWith(self, node: ast.AsyncWith) -> None:
        self.register_stmts(*node.body)

    def visit_Expr(self, node: ast.Expr) -> None:
        if isinstance(node.value, ast.Constant):
            return None

        if isinstance(node.value, ast.Call):
            return None

        if isinstance(node.value, ast.Lambda):
            return error.error("top-level lambdas must be named", culprit=node)

        name = node.__class__.__name__
        return error.error(f"unexpected top-level 'ast.{name}'", culprit=node)


def make_import_symbol(
    *,
    name: str,
    qualified_name: str,
    module_name: str,
    token: ast.Import | ast.ImportFrom,
) -> Import:
    """Return the import symbol for the given data, with error checking applied."""
    _import = Import(
        name=name,
        qualified_name=qualified_name,
        token=token,
    )

    # Don't require that blacklisted modules be locatable
    if is_in_import_blacklist(module_name):
        return _import

    # Ensure that we can follow the import
    if _import.origin is None:
        error.fatal(f"unable to find module {module_name!r}", culprit=token)

    return _import


def __dummy_token(name: Identifier) -> ast.Name:
    return ast.Name(
        id=name,
        lineno=0,
        col_offset=0,
        expr_context=ast.expr_context(),
    )


def __module_level_name(name: Identifier) -> Name:
    return Name(name=name, token=__dummy_token(name))


def __module_level_builtin(name: Identifier) -> Builtin:
    return Builtin(name=name, token=__dummy_token(name))


def error_starred_import_outside_init(
    node: ast.ImportFrom,
    module: ModuleName,
) -> None:
    code = gen_import_from_stmt(module, "*")
    message = f"do not use {code!r} outside of __init__.py files, be explicit"
    error.warning(message, culprit=node)



"""Assert that there are no assignment to names that identify imports."""
from __future__ import annotations

import ast
from contextlib import contextmanager
from typing import TYPE_CHECKING

from rattr.analyser.base import Assertor
from rattr.analyser.util import has_annotation
from rattr.ast.util import fullname_of, unravel_names
from rattr.models.context import Context
from rattr.models.symbol import CallInterface, Import

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable

    from rattr.ast.types import Identifier


def lhs_names(
    lhs: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
) -> list[str]:
    """Return the names in the lhs of the given assignment."""
    if hasattr(lhs, "targets"):
        targets = lhs.targets
    else:
        targets = [lhs.target]

    return [
        name
        for target in targets
        for name in unravel_names(target, _get_name=fullname_of)
    ]


def imports_in_node_lhs_names(
    node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
    *,
    context: Context,
) -> list[Import]:
    return imports_in_names(lhs_names(node), context=context)


def imports_in_names(
    names: Iterable[Identifier],
    *,
    context: Context,
) -> list[Import]:
    return [
        symbol
        for name in names
        if (symbol := context.get(name)) is not None
        if isinstance(symbol, Import)
    ]


class ImportClobberingAssertor(Assertor):
    def __init__(self, is_strict: bool = True) -> None:
        super().__init__(is_strict=is_strict)
        self.class_stack: list[str] = []

    @contextmanager
    def enter_class_name(self, name: str) -> Generator[None, None, None]:
        self.class_stack.append(name)
        yield
        self.class_stack.pop()

    def __clobbered(self, name: str, node: ast.AST) -> None:
        self.failed(f"redefinition of imported name {name!r}", culprit=node)

    def __deleted(self, name: str, node: ast.AST) -> None:
        self.failed(f"attempt to delete imported name {name!r}", culprit=node)

    def visit_Assign(self, node: ast.Assign) -> None:
        for import_ in imports_in_node_lhs_names(node, context=self.context):
            self.__clobbered(import_.name, node)

        return super().generic_visit(node)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        for import_ in imports_in_node_lhs_names(node, context=self.context):
            self.__clobbered(import_.name, node)

        return super().generic_visit(node)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        for import_ in imports_in_node_lhs_names(node, context=self.context):
            self.__clobbered(import_.name, node)

        return super().generic_visit(node)

    def visit_Delete(self, node: ast.Delete) -> None:
        for import_ in imports_in_node_lhs_names(node, context=self.context):
            self.__deleted(import_.name, node)

        return super().generic_visit(node)

    def visit_FunctionDef(self, node: ast.FunctionDef | ast.AsyncFunctionDef) -> None:
        if has_annotation("rattr_ignore", node):
            return

        if has_annotation("rattr_results", node):
            return

        is_in_method = self.class_stack != []

        if not is_in_method and isinstance(self.context.get(node.name), Import):
            return self.__clobbered(node.name, node)

        self.check_arguments(node)

        return super().generic_visit(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        return self.visit_FunctionDef(node)

    def visit_Lambda(self, node: ast.Lambda) -> None:
        self.check_arguments(node)
        return super().generic_visit(node)

    def check_arguments(
        self,
        node: ast.Lambda | ast.FunctionDef | ast.AsyncFunctionDef,
    ) -> None:
        interface = CallInterface.from_fn_def(node)
        names = {a for a in interface.all if a is not None}

        for import_ in imports_in_names(names, context=self.context):
            self.__clobbered(import_.name, node)

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        if has_annotation("rattr_ignore", node):
            return

        if has_annotation("rattr_results", node):
            return

        if isinstance(self.context.get(node.name), Import):
            self.__clobbered(node.name, node)
            return

        with self.enter_class_name(fullname_of(node, safe=True)):
            super().generic_visit(node)

    def visit_For(self, node: ast.For | ast.AsyncFor) -> None:
        names = [n for n in unravel_names(node.target)]

        for import_ in imports_in_names(names, context=self.context):
            self.__clobbered(import_.name, node)

        return super().generic_visit(node)

    def visit_AsyncFor(self, node: ast.AsyncFor) -> None:
        return self.visit_For(node)

    def visit_With(self, node: ast.With | ast.AsyncWith) -> None:
        for item in node.items:
            if item.optional_vars is None:
                continue

            names = [n for n in unravel_names(item.optional_vars)]

            for import_ in imports_in_names(names, context=self.context):
                self.__clobbered(import_.name, node)

        return super().generic_visit(node)

    def visit_AsyncWith(self, node: ast.AsyncWith) -> None:
        return self.visit_With(node)

    def check_comprehension(
        self,
        node: ast.ListComp | ast.SetComp | ast.GeneratorExp | ast.DictComp,
    ) -> None:
        for generator in node.generators:
            names = [n for n in unravel_names(generator.target)]

            for import_ in imports_in_names(names, context=self.context):
                self.__clobbered(import_.name, node)

        return super().generic_visit(node)

    def visit_ListComp(self, node: ast.ListComp) -> None:
        self.check_comprehension(node)

    def visit_SetComp(self, node: ast.SetComp) -> None:
        self.check_comprehension(node)

    def visit_GeneratorExp(self, node: ast.GeneratorExp) -> None:
        self.check_comprehension(node)

    def visit_DictComp(self, node: ast.DictComp) -> None:
        self.check_comprehension(node)



from __future__ import annotations

import pytest

from rattr.models.ir import FunctionIr
from rattr.models.symbol import Call, CallArguments, CallInterface, Func, Name
from rattr.results._simplify_utils import (
    construct_call_swaps,
    unbind_ir_with_call_swaps,
    unbind_name,
)
from tests.shared import match_output


@pytest.mark.parametrize(
    "name, basename, new_basename, resultant",
    [
        # Identity
        ("foo", "foo", "foo", "foo"),
        ("*foo", "foo", "foo", "*foo"),
        ("foo[]", "foo", "foo", "foo[]"),
        ("*foo[]", "foo", "foo", "*foo[]"),
        # Simple substitution
        ("foo", "foo", "bar", "bar"),
        ("*foo", "foo", "bar", "*bar"),
        ("foo[]", "foo", "bar", "bar[]"),
        ("*foo[]", "foo", "bar", "*bar[]"),
        # Complex identity
        (
            "comp.attr.m().res_attr[].item",
            "comp",
            "comp",
            "comp.attr.m().res_attr[].item",
        ),
        (
            "*comp.attr.m().res_attr[].item",
            "comp",
            "comp",
            "*comp.attr.m().res_attr[].item",
        ),
        (
            "comp.attr.m().res_attr[].item[]",
            "comp",
            "comp",
            "comp.attr.m().res_attr[].item[]",
        ),
        (
            "*comp.attr.m().res_attr[].item[]",
            "comp",
            "comp",
            "*comp.attr.m().res_attr[].item[]",
        ),
        # Complex substitution
        (
            "comp.attr.m().res_attr[].item",
            "comp",
            "bar",
            "bar.attr.m().res_attr[].item",
        ),
        (
            "*comp.attr.m().res_attr[].item",
            "comp",
            "bar",
            "*bar.attr.m().res_attr[].item",
        ),
        (
            "comp.attr.m().res_attr[].item[]",
            "comp",
            "bar",
            "bar.attr.m().res_attr[].item[]",
        ),
        (
            "*comp.attr.m().res_attr[].item[]",
            "comp",
            "bar",
            "*bar.attr.m().res_attr[].item[]",
        ),
    ],
)
def test_unbind_name(name: str, basename: str, new_basename: str, resultant: str):
    lhs = Name(name, basename)
    rhs = Name(resultant, new_basename)
    assert unbind_name(lhs, new_basename) == rhs


def test_unbind_ir_with_call_swaps_empty_ir_with_no_swaps():
    assert (
        unbind_ir_with_call_swaps(FunctionIr.the_empty_ir(), {})
        == FunctionIr.the_empty_ir()
    )


def test_unbind_ir_with_call_swaps_empty_ir_with_swaps():
    assert (
        unbind_ir_with_call_swaps(FunctionIr.the_empty_ir(), {"a": "b"})
        == FunctionIr.the_empty_ir()
    )


@pytest.fixture()
def example_func_ir_a():
    return {
        "sets": {
            Name("a"),
            Name("a.attr", "a"),
        },
        "gets": {
            Name("arg"),
            Name("arg.mth().res_attr[].value", "arg"),
        },
        "dels": {
            Name("bob"),
            Name("*dob", "dob"),
        },
        "calls": {
            Call(name="callee()", args=CallArguments(), target=None),
        },
    }


def test_unbind_ir_with_call_swaps_func_with_no_swaps(example_func_ir_a: FunctionIr):
    assert unbind_ir_with_call_swaps(example_func_ir_a, {}) == example_func_ir_a


def test_unbind_ir_with_call_swaps_func_with_simple_swap_single(
    example_func_ir_a: FunctionIr,
):
    swap = {"a": "b"}
    expected = {
        "sets": {
            Name("b"),
            Name("b.attr", "b"),
        },
        "gets": {
            Name("arg"),
            Name("arg.mth().res_attr[].value", "arg"),
        },
        "dels": {
            Name("bob"),
            Name("*dob", "dob"),
        },
        "calls": {
            Call(name="callee()", args=CallArguments(), target=None),
        },
    }
    assert unbind_ir_with_call_swaps(example_func_ir_a, swap) == expected


def test_unbind_ir_with_call_swaps_func_with_simple_swaps(
    example_func_ir_a: FunctionIr,
):
    swap = {
        "arg": "barg",
        "bob": "bobby",
    }
    expected = {
        "sets": {
            Name("a"),
            Name("a.attr", "a"),
        },
        "gets": {
            Name("barg"),
            Name("barg.mth().res_attr[].value", "barg"),
        },
        "dels": {
            Name("bobby"),
            Name("*dob", "dob"),
        },
        "calls": {
            Call(name="callee()", args=CallArguments(), target=None),
        },
    }
    assert unbind_ir_with_call_swaps(example_func_ir_a, swap) == expected


def test_unbind_ir_with_call_swaps_func_with_mixed_swaps(
    example_func_ir_a: FunctionIr,
):
    # Mixed insofar as some are relevant and some are irrelevant
    swap = {
        "dob": "dib",
        "xyz": "zyx",
    }
    expected = {
        "sets": {
            Name("a"),
            Name("a.attr", "a"),
        },
        "gets": {
            Name("arg"),
            Name("arg.mth().res_attr[].value", "arg"),
        },
        "dels": {
            Name("bob"),
            Name("*dib", "dib"),
        },
        "calls": {
            Call(name="callee()", args=CallArguments(), target=None),
        },
    }
    assert unbind_ir_with_call_swaps(example_func_ir_a, swap) == expected


def test_unbind_ir_with_call_swaps_regression():
    func_ir = FunctionIr.new(sets={Name(name="arg_b.set_in_fn_b", basename="arg_b")})
    expected = FunctionIr.new(sets={Name(name="arg.set_in_fn_b", basename="arg")})
    assert unbind_ir_with_call_swaps(func_ir, {"arg_b": "arg"}) == expected


@pytest.mark.parametrize(
    "interface, arguments, expected, expected_stderr",
    testcases := [
        # Empty
        (
            CallInterface(),
            CallArguments(),
            {},
            [],
        ),
        # Not enough args to satisfy positional arguments
        (
            CallInterface(posonlyargs=("a",)),
            CallArguments(args=()),
            {},
            [
                "call to 'fn' expected 1 posonlyargs but only received 0 "
                "positional arguments",
            ],
        ),
        (
            CallInterface(posonlyargs=("a",)),
            CallArguments(args=(), kwargs={"a": "fail?"}),
            {},
            [
                "call to 'fn' expected 1 posonlyargs but only received 0 "
                "positional arguments",
            ],
        ),
        # posonlyargs
        (
            CallInterface(args=("a", "b")),
            CallArguments(args=("a", "b")),
            {"a": "a", "b": "b"},
            [],
        ),
        (
            CallInterface(posonlyargs=("a_interface", "b_interface")),
            CallArguments(args=("a_call", "b_call"), kwargs={}),
            {"a_interface": "a_call", "b_interface": "b_call"},
            [],
        ),
        # positional args
        (
            CallInterface(posonlyargs=("a", "b"), args=("c", "d")),
            CallArguments(args=("A", "B", "C"), kwargs={"d": "D"}),
            {"a": "A", "b": "B", "c": "C", "d": "D"},
            [],
        ),
        (
            CallInterface(posonlyargs=("a", "b"), args=("c", "d")),
            CallArguments(args=("A", "B", "C"), kwargs={"d": "D", "e": "E"}),
            {"a": "A", "b": "B", "c": "C", "d": "D"},
            [
                "call to 'fn' received unexpected keyword arguments: ['e']",
            ],
        ),
        # positional args w/ vararg
        # ... with empty `rem`
        (
            CallInterface(posonlyargs=("a",), args=("b", "c"), vararg="rem"),
            CallArguments(args=("A", "B", "C"), kwargs={}),
            {"a": "A", "b": "B", "c": "C", "rem": "@Tuple"},
            [],
        ),
        # ... with non-empty `rem`
        (
            CallInterface(posonlyargs=("a",), args=("b", "c"), vararg="rem"),
            CallArguments(args=("A", "B", "C", "D", "E", "F"), kwargs={}),
            {"a": "A", "b": "B", "c": "C", "rem": "@Tuple"},
            [],
        ),
        # keyword args
        (
            CallInterface(args=("a", "b", "c")),
            CallArguments(args=("A",), kwargs={"c": "C", "b": "B"}),
            {"a": "A", "b": "B", "c": "C"},
            [],
        ),
        # ... with a vararg
        (
            CallInterface(args=("a", "b", "c"), vararg="rem"),
            CallArguments(args=("A", "B"), kwargs={"c": "C", "b": "B"}),
            {"a": "A", "b": "B", "c": "C", "rem": "@Tuple"},
            [
                "call to 'fn' received the arguments ['b'] by position and name",
            ],
        ),
        # kwonlyargs
        (
            CallInterface(kwonlyargs=("a", "b", "c")),
            CallArguments(args=("A", "B"), kwargs={"c": "C"}),
            {"c": "C"},
            [
                "call to 'fn' received too many positional arguments",
            ],
        ),
        (
            CallInterface(kwonlyargs=("a", "b", "c")),
            CallArguments(args=(), kwargs={"c": "C"}),
            {"c": "C"},
            [],
        ),
        (
            CallInterface(kwonlyargs=("a", "b", "c")),
            CallArguments(args=(), kwargs={"a": "A", "c": "C", "b": "B"}),
            {"a": "A", "b": "B", "c": "C"},
            [],
        ),
        # keyword args
        (
            CallInterface(args=("a",), vararg="rem", kwonlyargs=("b", "c")),
            CallArguments(args=("A", "extra"), kwargs={"c": "C", "b": "B"}),
            {"a": "A", "b": "B", "c": "C", "rem": "@Tuple"},
            [],
        ),
        # ... with assumed default
        (
            CallInterface(args=("a", "b", "c", "d", "e")),
            CallArguments(args=("not_a", "not_b"), kwargs={"d": "not_d", "c": "not_c"}),
            {"a": "not_a", "b": "not_b", "c": "not_c", "d": "not_d"},
            [],
        ),
        # kwarg
        (
            CallInterface(
                args=("a",),
                vararg="rem",
                kwonlyargs=("b", "c"),
                kwarg="kwargs",
            ),
            CallArguments(args=("A", "extra"), kwargs={"c": "C", "b": "B"}),
            {"a": "A", "b": "B", "c": "C", "rem": "@Tuple"},
            [],
        ),
        (
            CallInterface(
                args=("a",),
                vararg="rem",
                kwonlyargs=("b", "c"),
                kwarg="kwargs",
            ),
            CallArguments(
                args=("A", "extra"),
                kwargs={
                    "c": "C",
                    "b": "B",
                    "kw": "a_kwarg",
                    "kw2": "another_kwarg",
                },
            ),
            {"a": "A", "b": "B", "c": "C", "rem": "@Tuple", "kwargs": "@Dict"},
            [],
        ),
        # Complex examples
        (
            CallInterface(
                posonlyargs=("a", "b"),
                args=("c", "d"),
                vararg="args",
                kwonlyargs=("e",),
                kwarg="kwarg",
            ),
            CallArguments(args=(), kwargs={}),
            {},
            [
                "call to 'fn' expected 2 posonlyargs but only received 0 "
                "positional arguments",
            ],
        ),
        (
            CallInterface(
                posonlyargs=("a", "b"),
                args=("c", "d"),
                vararg="args",
                kwonlyargs=("e",),
                kwarg="kwarg",
            ),
            CallArguments(args=("A", "B", "C", "D", "extra"), kwargs={"e": "E"}),
            {"a": "A", "b": "B", "c": "C", "d": "D", "args": "@Tuple", "e": "E"},
            [],
        ),
        (
            CallInterface(
                posonlyargs=("a", "b"),
                args=("c", "d"),
                vararg="args",
                kwonlyargs=("e",),
                kwarg="kwarg",
            ),
            CallArguments(args=("A", "B", "C", "D", "extra"), kwargs={}),
            {"a": "A", "b": "B", "c": "C", "d": "D", "args": "@Tuple"},
            [],
        ),
        (
            CallInterface(
                posonlyargs=("a", "b"),
                args=("c", "d"),
                vararg="args",
                kwonlyargs=("e",),
                kwarg="kwarg",
            ),
            CallArguments(args=("A", "B", "C"), kwargs={"e": "E", "d": "D"}),
            {"a": "A", "b": "B", "c": "C", "d": "D", "args": "@Tuple", "e": "E"},
            [],
        ),
        # Some errors
        (
            CallInterface(args=("a",)),
            CallArguments(args=("not_a",), kwargs={"a": "not_a"}),
            {"a": "not_a"},
            ["call to 'fn' received the arguments ['a'] by position and name"],
        ),
        (
            CallInterface(args=("a", "b")),
            CallArguments(args=(), kwargs={"c": "no_match", "d": "no_match"}),
            {},
            ["call to 'fn' received unexpected keyword arguments: ['d', 'c']"],
        ),
    ],
    ids=[i for i, _ in enumerate(testcases)],
)
def test_construct_swap(
    interface: CallInterface,
    arguments: CallArguments,
    expected: dict[str, str],
    expected_stderr: list[str],
    capfd: pytest.CaptureFixture[str],
):
    fn_def = Func(name="fn", interface=interface)
    fn_call = Call(name="fn", args=arguments)

    assert construct_call_swaps(fn_def, fn_call) == expected

    _, stderr = capfd.readouterr()
    assert match_output(stderr, expected_stderr)



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

import pytest

from rattr.models.ir import FileIr
from rattr.models.symbol import (
    Builtin,
    Call,
    CallArguments,
    CallInterface,
    Class,
    Func,
    Import,
)
from rattr.results._find_call_target import find_call_target_and_ir
from rattr.results._types import IrCall, IrEnvironment, IrTarget
from tests.shared import Import_, match_output

if TYPE_CHECKING:
    from typing import Literal

    from tests.shared import FileIrFromDictFn


@pytest.fixture()
def dummy_caller() -> mock.Mock:
    return mock.Mock()


def test_find_call_target_and_ir_target_is_not_defined_in_env(
    dummy_caller: mock.Mock,
    example_environment_a: IrEnvironment,
):
    call = IrCall(
        caller=dummy_caller,
        symbol=Call(name="some_func()", args=CallArguments(), target=None),
    )
    assert find_call_target_and_ir(call, environment=example_environment_a) is None


def test_find_call_target_and_ir_target_is_a_builtin(
    dummy_caller: mock.Mock,
    example_environment_a: IrEnvironment,
):
    call = IrCall(
        caller=dummy_caller,
        symbol=Call(name="max()", args=CallArguments(), target=Builtin("max")),
    )
    assert find_call_target_and_ir(call, environment=example_environment_a) is None


@pytest.mark.parametrize("fn", ["a", "b"])
def test_find_call_target_and_ir_target_is_func_defined_in_env(
    dummy_caller: mock.Mock,
    example_environment_a: IrEnvironment,
    example_file_ir_a: FileIr,
    fn: Literal["a", "b"],
):
    fn_symbol = Func(name=f"fn_{fn}", interface=CallInterface())
    fn_call = IrCall(
        caller=dummy_caller,
        symbol=Call(name=f"fn_{fn}()", args=CallArguments(), target=fn_symbol),
    )
    fn_ir = example_file_ir_a[fn_symbol]

    actual = find_call_target_and_ir(fn_call, environment=example_environment_a)
    expected = IrTarget(symbol=fn_symbol, ir=fn_ir)

    assert actual == expected


def test_find_call_target_and_ir_target_is_a_class_init_defined_in_env(
    dummy_caller: mock.Mock,
    example_environment_a: IrEnvironment,
    example_file_ir_a: FileIr,
):
    cls_symbol = Class(name="ClassA", interface=CallInterface(args=("self", "arg")))
    cls_call = IrCall(
        caller=dummy_caller,
        symbol=Call(
            name="ClassA()",
            args=CallArguments(args=("a",)),
            target=cls_symbol,
        ),
    )
    cls_ir = example_file_ir_a[cls_symbol]

    actual = find_call_target_and_ir(cls_call, environment=example_environment_a)
    expected = IrTarget(symbol=cls_symbol, ir=cls_ir)

    assert actual == expected


@pytest.mark.pypy()
def test_find_call_target_and_ir_target_is_in_stdlib(
    example_environment_a: IrEnvironment,
):
    stdlib_call_symbol = Call(
        name="sin()",
        args=CallArguments(),
        target=Import("sin", "math.sin"),
    )
    assert stdlib_call_symbol.target.module_name == "math"

    actual = find_call_target_and_ir(
        IrCall(
            mock.Mock(),  # unused when symbol is an import
            symbol=stdlib_call_symbol,
        ),
        environment=example_environment_a,
    )
    assert actual is None


def test_find_call_target_and_ir_target_is_imported(
    dummy_caller: mock.Mock,
    example_file_ir_a: FileIr,
    file_ir_from_dict: FileIrFromDictFn,
):
    fn = Func(name="fn", interface=CallInterface())
    fn_ir = {
        "sets": set(),
        "gets": set(),
        "dels": set(),
        "calls": set(),
    }
    expected = IrTarget(symbol=fn, ir=fn_ir)

    environment = IrEnvironment(
        target_ir=example_file_ir_a,
        import_irs={"module": file_ir_from_dict({fn: fn_ir})},
    )

    call = IrCall(
        caller=dummy_caller,
        symbol=Call(
            name="fn()",
            args=CallArguments(),
            target=Import_(
                "fn",
                "module.fn",
                module_name_and_spec=("module", mock.Mock()),
            ),
        ),
    )

    assert find_call_target_and_ir(call, environment=environment) == expected


def test_find_call_target_and_ir_target_is_import_undefined_in_module(
    dummy_caller: mock.Mock,
    example_file_ir_a: FileIr,
    file_ir_from_dict: FileIrFromDictFn,
    capfd: pytest.CaptureFixture[str],
):
    fn = Func(name="fn", interface=CallInterface())
    fn_ir = {
        "sets": set(),
        "gets": set(),
        "dels": set(),
        "calls": set(),
    }

    environment = IrEnvironment(
        target_ir=example_file_ir_a,
        import_irs={"module": file_ir_from_dict({fn: fn_ir})},
    )

    call = IrCall(
        caller=dummy_caller,
        symbol=Call(
            name="nope()",
            args=CallArguments(),
            target=Import_(
                "nope",
                "module.nope",
                module_name_and_spec=("module", mock.Mock()),
            ),
        ),
    )
    assert find_call_target_and_ir(call, environment=environment) is None

    _, stderr = capfd.readouterr()
    assert match_output(
        stderr,
        [
            "unable to resolve call to 'nope' in import 'module', "
            "it is likely undefined",
        ],
    )


def test_find_call_target_and_ir_target_is_import_in_undefined_module(
    dummy_caller: mock.Mock,
    example_file_ir_a: FileIr,
    file_ir_from_dict: FileIrFromDictFn,
):
    fn = Func(name="fn", interface=CallInterface())
    fn_ir = {
        "sets": set(),
        "gets": set(),
        "dels": set(),
        "calls": set(),
    }

    environment = IrEnvironment(
        target_ir=example_file_ir_a,
        import_irs={"module": file_ir_from_dict({fn: fn_ir})},
    )

    call = IrCall(
        caller=dummy_caller,
        symbol=Call(
            name="nah()",
            args=CallArguments(),
            target=Import_(
                "nah",
                "noway.nah",
                module_name_and_spec=("noway", mock.Mock()),
            ),
        ),
    )
    with pytest.raises(ImportError):
        assert find_call_target_and_ir(call, environment=environment) is None



from __future__ import annotations

import ast
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from rattr.analyser.file import FileAnalyser
from rattr.models.context import Context, compile_root_context
from rattr.models.ir import FileIr
from rattr.models.symbol import (
    AnyCallInterface,
    Call,
    CallArguments,
    CallInterface,
    Class,
    Func,
    Import,
    Name,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from tests.shared import MakeSymbolTableFn, ParseFn, StateFn


@pytest.fixture(autouse=True)
def __set_current_file(state: StateFn) -> Iterator[None]:
    with state(current_file=Path(__file__)):
        yield


class TestClassAnalyser:
    def test_class_attributes(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        ast_ = parse(
            """
            class Numbers(NotARealEnum):
                One = 1
                Two = 2
                Three = 3
                Four, Five = 4, 5
            """
        )
        file_analyser = FileAnalyser(ast_, compile_root_context(ast_))
        file_analyser.analyse()

        expected = make_symbol_table(
            [
                Class(name="Numbers", interface=AnyCallInterface()),
                Name(name="Numbers.One", basename="Numbers"),
                Name(name="Numbers.Two", basename="Numbers"),
                Name(name="Numbers.Three", basename="Numbers"),
                Name(name="Numbers.Four", basename="Numbers"),
                Name(name="Numbers.Five", basename="Numbers"),
            ],
            include_root_symbols=True,
        )

        assert file_analyser.context.symbol_table == expected

    def test_class_initialiser(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        ast_ = parse(
            """
            class SomeClass:
                def __init__(self, some_arg):
                    self.whatever = some_arg.whatever
                    some_arg.good_number = 5    # 5 is a good number
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        some_class_symbol = Class(
            name="SomeClass",
            interface=CallInterface(args=("self", "some_arg")),
        )

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [some_class_symbol], include_root_symbols=True
                ),
            ),
            file_ir={
                some_class_symbol: {
                    "sets": {
                        Name("self.whatever", "self"),
                        Name("some_arg.good_number", "some_arg"),
                    },
                    "gets": {
                        Name("some_arg.whatever", "some_arg"),
                    },
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected

    def test_func_calls_class_initialiser(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        ast_ = parse(
            """
            def func(arg):
                return SomeClass(arg)

            class SomeClass:
                def __init__(self, some_arg):
                    self.whatever = some_arg.whatever
                    some_arg.good_number = 5    # 5 is a good number
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        func = Func(name="func", interface=CallInterface(args=("arg",)))
        cls = Class(
            name="SomeClass",
            interface=CallInterface(args=("self", "some_arg")),
        )
        cls_init_call = Call(
            name="SomeClass",
            args=CallArguments(
                args=(
                    "@ReturnValue",
                    "arg",
                ),
                kwargs={},
            ),
            target=cls,
        )

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table([func, cls], include_root_symbols=True),
            ),
            file_ir={
                func: {
                    "gets": {Name("arg")},
                    "sets": set(),
                    "dels": set(),
                    "calls": {cls_init_call},
                },
                cls: {
                    "gets": {
                        Name("some_arg.whatever", "some_arg"),
                    },
                    "sets": {
                        Name("self.whatever", "self"),
                        Name("some_arg.good_number", "some_arg"),
                    },
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_class_initialiser_calls_func(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        ast_ = parse(
            """
            def func(fn_arg):
                return fn_arg.some_attr

            class SomeClass:
                def __init__(self, some_arg):
                    self.a_thing = func(some_arg)
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        func = Func(name="func", interface=CallInterface(args=("fn_arg",)))
        cls = Class(
            name="SomeClass",
            interface=CallInterface(args=("self", "some_arg")),
        )
        func_call = Call(
            name="func",
            args=CallArguments(
                args=("some_arg",),
                kwargs={},
            ),
            target=func,
        )

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table([func, cls], include_root_symbols=True),
            ),
            file_ir={
                func: {
                    "sets": set(),
                    "gets": {Name("fn_arg.some_attr", "fn_arg")},
                    "dels": set(),
                    "calls": set(),
                },
                cls: {
                    "sets": {
                        Name("self.a_thing", "self"),
                    },
                    "gets": {
                        Name("some_arg"),
                    },
                    "dels": set(),
                    "calls": {func_call},
                },
            },
        )

        assert results == expected

    def test_staticmethod(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        ast_ = parse(
            """
            class SomeClass:
                @staticmethod
                def method(a, b):
                    a.attr = b.attr
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        cls = Class(name="SomeClass", interface=AnyCallInterface())
        method = Func(name="SomeClass.method", interface=CallInterface(args=("a", "b")))

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [
                        cls,
                        method,
                    ],
                    include_root_symbols=True,
                ),
            ),
            file_ir={
                method: {
                    "sets": {Name("a.attr", "a")},
                    "gets": {Name("b.attr", "b")},
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected

    def test_staticmethod_and_init(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        # Static method in "mixed" class
        ast_ = parse(
            """
            class SomeClass:
                def __init__(self, arg):
                    self.attr = arg

                @staticmethod
                def method(a, b):
                    a.attr = b.attr
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        cls = Class(name="SomeClass", interface=CallInterface(args=("self", "arg")))
        method = Func(name="SomeClass.method", interface=CallInterface(args=("a", "b")))

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [
                        cls,
                        method,
                    ],
                    include_root_symbols=True,
                ),
            ),
            file_ir={
                cls: {
                    "sets": {Name("self.attr", "self")},
                    "gets": {Name("arg")},
                    "dels": set(),
                    "calls": set(),
                },
                method: {
                    "sets": {Name("a.attr", "a")},
                    "gets": {Name("b.attr", "b")},
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_class_with_static_members(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        ast_ = parse(
            """
            class FibIter:
                N  = 0
                N_ = 1

                @staticmethod
                def next() -> int:
                    last = FibIter.N

                    FibIter.N  = FibIter.N_
                    FibIter.N_ = FibIter.N_ + last

                    return FibIter.N
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        cls = Class(name="FibIter", interface=AnyCallInterface())
        cls_n = Name(name="FibIter.N", basename="FibIter")
        cls_n_ = Name(name="FibIter.N_", basename="FibIter")
        cls_next = Func(name="FibIter.next", interface=CallInterface(args=()))

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [
                        cls,
                        cls_n,
                        cls_n_,
                        cls_next,
                    ],
                    include_root_symbols=True,
                ),
            ),
            file_ir={
                cls_next: {
                    "sets": {
                        Name("last"),
                        Name("FibIter.N", "FibIter"),
                        Name("FibIter.N_", "FibIter"),
                    },
                    "gets": {
                        Name("last"),
                        Name("FibIter.N", "FibIter"),
                        Name("FibIter.N_", "FibIter"),
                    },
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected

    def test_enum_by_explicit_import(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        ast_ = parse(
            """
            from enum import Enum

            class MyEnum(Enum):
                LHS = "RHS"
            """
        )
        context = compile_root_context(ast_).expand_starred_imports()
        results = FileAnalyser(ast_, context).analyse()

        enum_import = Import(name="Enum", qualified_name="enum.Enum")
        my_enum = Class(name="MyEnum", interface=CallInterface(args=("self", "_id")))
        my_enum_lhs = Name(name="MyEnum.LHS", basename="MyEnum")

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [
                        enum_import,
                        my_enum,
                        my_enum_lhs,
                    ],
                    include_root_symbols=True,
                ),
            ),
            file_ir={
                my_enum: {
                    "sets": set(),
                    "gets": {Name("MyEnum.LHS", "MyEnum")},
                    "calls": set(),
                    "dels": set(),
                }
            },
        )

        assert results == expected

    def test_enum_by_module_import(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        ast_ = parse(
            """
            import enum

            class MyEnum(enum.Enum):
                LHS = "RHS"
            """
        )
        context = compile_root_context(ast_).expand_starred_imports()
        results = FileAnalyser(ast_, context).analyse()

        enum_import = Import(name="enum", qualified_name="enum")
        my_enum = Class(name="MyEnum", interface=CallInterface(args=("self", "_id")))
        my_enum_lhs = Name(name="MyEnum.LHS", basename="MyEnum")

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [
                        enum_import,
                        my_enum,
                        my_enum_lhs,
                    ],
                    include_root_symbols=True,
                ),
            ),
            file_ir={
                my_enum: {
                    "sets": set(),
                    "gets": {Name("MyEnum.LHS", "MyEnum")},
                    "calls": set(),
                    "dels": set(),
                }
            },
        )

        assert results == expected

    def test_enum_by_starred_import(self, parse: ParseFn):
        ast_ = parse(
            """
            from enum import *

            class MyEnum(Enum):
                LHS = "RHS"
            """
        )
        context = compile_root_context(ast_).expand_starred_imports()
        results = FileAnalyser(ast_, context).analyse()

        my_enum = Class(name="MyEnum", interface=CallInterface(args=("self", "_id")))
        my_enum_lhs = Name(name="MyEnum.LHS", basename="MyEnum")

        expected_context = compile_root_context(ast.parse("from enum import *"))
        expected_context = expected_context.expand_starred_imports()
        expected_context.add(my_enum)
        expected_context.add(my_enum_lhs)

        expected = FileIr(
            context=expected_context,
            file_ir={
                my_enum: {
                    "sets": set(),
                    "gets": {Name("MyEnum.LHS", "MyEnum")},
                    "calls": set(),
                    "dels": set(),
                }
            },
        )

        assert results == expected

    def test_enum_call(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
        constant: str,
    ):
        ast_ = parse(
            """
            from enum import Enum

            class MyEnum(Enum):
                ONE = "one"
                TWO = "two"
                THREE = "three"

            def get_one():
                # actually gets all!
                return MyEnum("one")

            def get_none():
                return 4
            """
        )
        context = compile_root_context(ast_).expand_starred_imports()
        results = FileAnalyser(ast_, context).analyse()

        enum_import = Import(name="Enum", qualified_name="enum.Enum")
        get_one = Func(name="get_one", interface=CallInterface())
        get_none = Func(name="get_none", interface=CallInterface())

        my_enum = Class(name="MyEnum", interface=CallInterface(args=("self", "_id")))
        my_enum_one = Name(name="MyEnum.ONE", basename="MyEnum")
        my_enum_two = Name(name="MyEnum.TWO", basename="MyEnum")
        my_enum_three = Name(name="MyEnum.THREE", basename="MyEnum")
        enum_call = Call(
            name="MyEnum()",
            args=CallArguments(args=("@ReturnValue", constant)),
            target=my_enum,
        )

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [
                        enum_import,
                        get_one,
                        get_none,
                        my_enum,
                        my_enum_one,
                        my_enum_two,
                        my_enum_three,
                    ],
                    include_root_symbols=True,
                ),
            ),
            file_ir={
                my_enum: {
                    "sets": set(),
                    "gets": {
                        Name("MyEnum.ONE", "MyEnum"),
                        Name("MyEnum.TWO", "MyEnum"),
                        Name("MyEnum.THREE", "MyEnum"),
                    },
                    "calls": set(),
                    "dels": set(),
                },
                get_one: {
                    "sets": set(),
                    "gets": set(),
                    "calls": {enum_call},
                    "dels": set(),
                },
                get_none: {
                    "sets": set(),
                    "gets": set(),
                    "calls": set(),
                    "dels": set(),
                },
            },
        )

        assert results == expected

    def test_walrus_in_class_static_member(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        ast_ = parse(
            """
            class IAmTheWalrus:
                cls_attr_a = (cls_attr_b := 0)  # Yuck

                def __init__(self):
                    inst_attr_a = (inst_attr_b := 0)  # Just as yuck
            """
        )
        file_analyser = FileAnalyser(ast_, compile_root_context(ast_))
        results = file_analyser.analyse()

        wally = Class(name="IAmTheWalrus", interface=CallInterface(args=("self",)))
        cls_attr_a = Name(name="IAmTheWalrus.cls_attr_a", basename="IAmTheWalrus")
        cls_attr_b = Name(name="IAmTheWalrus.cls_attr_b", basename="IAmTheWalrus")

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [
                        wally,
                        cls_attr_a,
                        cls_attr_b,
                    ],
                    include_root_symbols=True,
                ),
            ),
            file_ir={
                wally: {
                    "sets": {
                        Name("inst_attr_a"),
                        Name("inst_attr_b"),
                    },
                    "gets": set(),
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected

    def test_walrus_in_class_static_member_in_tuple(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        ast_ = parse(
            """
            class IAmTheWalrus:
                cls_attr_a = (a, cls_attr_b := 0)

                def __init__(self):
                    inst_attr_a = (a, inst_attr_b := 0)
            """
        )
        file_analyser = FileAnalyser(ast_, compile_root_context(ast_))
        results = file_analyser.analyse()

        wally = Class(name="IAmTheWalrus", interface=CallInterface(args=("self",)))
        cls_attr_a = Name(name="IAmTheWalrus.cls_attr_a", basename="IAmTheWalrus")
        cls_attr_b = Name(name="IAmTheWalrus.cls_attr_b", basename="IAmTheWalrus")

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [
                        wally,
                        cls_attr_a,
                        cls_attr_b,
                    ],
                    include_root_symbols=True,
                ),
            ),
            file_ir={
                wally: {
                    "sets": {
                        Name("inst_attr_a"),
                        Name("inst_attr_b"),
                    },
                    "gets": {
                        Name("a"),
                    },
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected

    def test_walrus_method(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        ast_ = parse(
            """
            class IAmTheWalrus:
                def contrived(self):
                    this_is = (a_very_contrived := self.operation)

                @staticmethod
                def more_contrived():
                    if (alias := this_is_just_so_long_i_will_alias_it):
                        return alias
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        wally = Class(name="IAmTheWalrus", interface=AnyCallInterface())
        static = Func(name="IAmTheWalrus.more_contrived", interface=CallInterface())

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [
                        wally,
                        static,
                    ],
                    include_root_symbols=True,
                ),
            ),
            file_ir={
                static: {
                    "sets": {
                        Name("alias"),
                    },
                    "gets": {
                        Name("this_is_just_so_long_i_will_alias_it"),
                        Name("alias"),
                    },
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected



from __future__ import annotations

from rattr.versioning._util import is_python_version

if is_python_version(">=3.10"):
    from typing import ParamSpec, TypeAlias  # noqa: F401
else:
    from typing_extensions import ParamSpec, TypeAlias  # noqa: F401

if is_python_version(">=3.11"):
    from typing import Self  # noqa: F401
else:
    from typing_extensions import Self  # noqa: F401



from __future__ import annotations

from pathlib import Path

from setuptools import find_packages, setup

HERE = Path(__file__).parent.resolve()
README = HERE / "README.md"
REQUIREMENTS = HERE / "requirements.txt"


setup(
    name="rattr",
    use_scm_version={
        "write_to": "rattr/_version.py",
        "write_to_template": (
            'version = "{version}"  # this should be overwritten by setuptools_scm\n'
        ),
    },
    setup_requires=["setuptools_scm"],
    author="Suade Labs, Brandon Harris",
    author_email="brandon@saude.org, bpharris@pm.me",
    maintainer="Brandon Harris",
    maintainer_email="brandon@saude.org, bpharris@pm.me",
    packages=find_packages(),
    description="Rattr rats on your attrs.",
    long_description=README.read_text(),
    long_description_content_type="text/markdown",
    install_requires=[r.strip() for r in REQUIREMENTS.read_text().splitlines()],
    extras_require={
        "dev": [
            "pytest>=8.0.0",
            "ruff==0.2.1",
        ],
    },
    entry_points={"console_scripts": ["rattr = rattr.__main__:entry_point"]},
    keywords="automation linting type-checking attributes rats",
    url="https://github.com/SuadeLabs/rattr",
    license="MIT",
    python_requires=">=3.9",
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Quality Assurance",
    ],
)



from rattr.analyser.annotations import rattr_ignore, rattr_results

from .module.foo import my_lib_function


def my_foo(x):
    return x.attr


def some_function(x, y, z):
    x.some_thing = my_foo(x) + my_foo(y) + my_foo(z)

    if x.is_whatever:
        x.another_thing = x.whatever + 10
    else:
        x.another_thing = 0

    return x


@rattr_ignore()
def this_should_not_appear_in_the_results(foo):
    return foo.bar + foo.baz


@rattr_results(
    gets={
        "foobar.x",
        "foobar.y",
        "foobar.z",
    },
    sets={
        "some_other_var.attr",
    },
    calls=[
        ("this_should_not_appear_in_the_results()", (["foobar"], {})),
        ("some_function()", ([], {"a": "@Int", "b": "@Int", "c": "foobar"})),
    ],
)
def overwrite_me(foobar):
    print(foobar.baz)


def i_use_a_lib_function(data):
    good_data = [datum for datum in data if my_lib_function(datum)]

    if len(good_data):
        raise Exception(f"problem with {data[0].goodness}, ...")

    return good_data



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from rattr.models.symbol import Call, CallArguments, CallInterface, Class, Func
from rattr.results._types import IrCall, IrCallTreeNode, IrTarget
from rattr.results.util import make_target_ir_call_tree
from tests.shared import match_output

if TYPE_CHECKING:
    from rattr.results._types import IrEnvironment


def test_make_target_ir_call_tree_a(
    example_environment_a: IrEnvironment,
):
    # Each function in example_environment_a has no calls, so each call tree is only the
    # root node
    file_ir = example_environment_a.target_ir
    symbols = (
        Func(name="fn_a", interface=CallInterface()),
        Func(name="fn_b", interface=CallInterface()),
        Class(name="ClassA", interface=CallInterface(args=("self", "arg"))),
    )
    for symbol in symbols:
        target = IrTarget(symbol=symbol, ir=file_ir[symbol])
        expected = IrCallTreeNode(
            target=target,
            edge_in=None,
            edges_out=[],
            children=[],
        )
        assert (
            make_target_ir_call_tree(target, environment=example_environment_a)
            == expected
        )


def test_make_target_ir_call_tree_b(
    example_environment_b: IrEnvironment,
):
    file_ir = example_environment_b.target_ir

    fn_d = Func(name="fn_d", interface=CallInterface(args=("d",)))
    fn_d_target = IrTarget(symbol=fn_d, ir=file_ir[fn_d])
    fn_d_actual = make_target_ir_call_tree(
        fn_d_target,
        environment=example_environment_b,
    )
    fn_d_expected = IrCallTreeNode(
        target=fn_d_target,
        edge_in=None,
        edges_out=[],
        children=[],
    )
    assert fn_d_actual == fn_d_expected

    fn_c = Func(name="fn_c", interface=CallInterface(args=("c",)))
    fn_c_target = IrTarget(symbol=fn_c, ir=file_ir[fn_c])
    fn_c_actual = make_target_ir_call_tree(
        fn_c_target,
        environment=example_environment_b,
    )
    fn_c_expected = IrCallTreeNode(
        target=fn_c_target,
        edge_in=None,
        edges_out=[],
        children=[],
    )
    assert fn_c_actual == fn_c_expected

    fn_b = Func(name="fn_b", interface=CallInterface(args=("b",)))
    fn_b_target = IrTarget(symbol=fn_b, ir=file_ir[fn_b])
    fn_b_actual = make_target_ir_call_tree(
        fn_b_target,
        environment=example_environment_b,
    )
    fn_b_expected = IrCallTreeNode(
        target=fn_b_target,
        edge_in=None,
        edges_out=[
            fn_b_to_fn_c_call := IrCall(
                caller=fn_b,
                symbol=Call(name="fn_c", args=CallArguments(args=("b",)), target=fn_c),
            ),
            fn_b_to_fn_d_call := IrCall(
                caller=fn_b,
                symbol=Call(name="fn_d", args=CallArguments(args=("b",)), target=fn_d),
            ),
        ],
        children=[
            IrCallTreeNode(
                target=fn_c_expected.target,
                edge_in=fn_b_to_fn_c_call,
                edges_out=fn_c_expected.edges_out,
                children=fn_c_expected.children,
            ),
            IrCallTreeNode(
                target=fn_d_expected.target,
                edge_in=fn_b_to_fn_d_call,
                edges_out=fn_d_expected.edges_out,
                children=fn_d_expected.children,
            ),
        ],
    )
    assert fn_b_actual == fn_b_expected

    fn_a = Func(name="fn_a", interface=CallInterface(args=("a",)))
    fn_a_target = IrTarget(symbol=fn_a, ir=file_ir[fn_a])
    fn_a_actual = make_target_ir_call_tree(
        fn_a_target,
        environment=example_environment_b,
    )
    fn_a_expected = IrCallTreeNode(
        target=fn_a_target,
        edge_in=None,
        edges_out=[
            fn_a_to_fn_b_call := IrCall(
                caller=fn_a,
                symbol=Call(name="fn_b", args=CallArguments(args=("a",)), target=fn_b),
            ),
        ],
        children=[
            IrCallTreeNode(
                target=fn_b_expected.target,
                edge_in=fn_a_to_fn_b_call,
                edges_out=fn_b_expected.edges_out,
                children=fn_b_expected.children,
            ),
        ],
    )
    assert fn_a_actual == fn_a_expected


def test_make_target_ir_call_tree_c(
    example_environment_c: IrEnvironment,
    capfd: pytest.CaptureFixture[str],
):
    file_ir = example_environment_c.target_ir

    fn_b = Func(name="fn_b", interface=CallInterface(args=("b",)))
    fn_b_target = IrTarget(symbol=fn_b, ir=file_ir[fn_b])
    fn_b_actual = make_target_ir_call_tree(
        fn_b_target,
        environment=example_environment_c,
    )
    fn_b_expected = IrCallTreeNode(
        target=fn_b_target,
        edge_in=None,
        edges_out=[
            IrCall(
                caller=fn_b,
                symbol=Call(
                    name="some_other_undefined_func",
                    args=CallArguments(args=()),
                ),
            ),
        ],
        children=[],
    )
    assert fn_b_actual == fn_b_expected

    fn_a = Func(name="fn_a", interface=CallInterface(args=("a",)))
    fn_a_target = IrTarget(symbol=fn_a, ir=file_ir[fn_a])
    fn_a_actual = make_target_ir_call_tree(
        fn_a_target,
        environment=example_environment_c,
    )
    fn_a_expected = IrCallTreeNode(
        target=fn_a_target,
        edge_in=None,
        edges_out=[
            fn_a_to_fn_b_call := IrCall(
                caller=fn_a,
                symbol=Call(name="fn_b", args=CallArguments(args=("a",)), target=fn_b),
            ),
            IrCall(
                caller=fn_a,
                symbol=Call(
                    name="some_undefined_func",
                    args=CallArguments(args=()),
                ),
            ),
        ],
        children=[
            IrCallTreeNode(
                target=fn_b_expected.target,
                edge_in=fn_a_to_fn_b_call,
                edges_out=fn_b_expected.edges_out,
                children=fn_b_expected.children,
            ),
        ],
    )
    assert fn_a_actual == fn_a_expected

    _, stderr = capfd.readouterr()
    assert match_output(stderr, [])


def test_make_target_ir_call_tree_d(
    example_environment_d: IrEnvironment,
    capfd: pytest.CaptureFixture[str],
):
    file_ir = example_environment_d.target_ir

    fn_a = Func(name="fn_a", interface=CallInterface(args=("a",)))
    fn_a_target = IrTarget(symbol=fn_a, ir=file_ir[fn_a])
    fn_a_actual = make_target_ir_call_tree(
        fn_a_target,
        environment=example_environment_d,
    )
    fn_a_expected = IrCallTreeNode(
        target=fn_a_target,
        edge_in=None,
        edges_out=[
            IrCall(
                caller=fn_a,
                symbol=Call("math.max", args=CallArguments(args=())),
            ),
            IrCall(
                caller=fn_a,
                symbol=Call("os.path.join", args=CallArguments(args=())),
            ),
        ],
        children=[],
    )
    assert fn_a_actual == fn_a_expected

    _, stderr = capfd.readouterr()
    assert match_output(stderr, [])


def test_make_target_ir_call_tree_e(
    example_environment_e: IrEnvironment,
    capfd: pytest.CaptureFixture[str],
):
    file_ir = example_environment_e.target_ir

    fn_a = Func(name="fn_a", interface=CallInterface(args=("a",)))
    fn_a_target = IrTarget(symbol=fn_a, ir=file_ir[fn_a])
    fn_a_actual = make_target_ir_call_tree(
        fn_a_target,
        environment=example_environment_e,
    )
    fn_a_expected = IrCallTreeNode(
        target=fn_a_target,
        edge_in=None,
        edges_out=[
            IrCall(
                caller=fn_a,
                symbol=Call("enumerate", args=CallArguments(args=())),
            ),
            IrCall(
                caller=fn_a,
                symbol=Call("max", args=CallArguments(args=())),
            ),
        ],
        children=[],
    )
    assert fn_a_actual == fn_a_expected

    _, stderr = capfd.readouterr()
    assert match_output(stderr, [])


def test_make_target_ir_call_tree_f(
    example_environment_f: IrEnvironment,
    capfd: pytest.CaptureFixture[str],
):
    # This test environment has lots of direct and indirect recursion, which is handled
    # correctly with the cycles removed but remains quite long.
    file_ir = example_environment_f.target_ir

    fn_a_target = IrTarget(
        symbol=(fn_a := Func(name="fn_a", interface=CallInterface(args=("a",)))),
        ir=file_ir[fn_a],
    )
    fn_b_target = IrTarget(
        symbol=(fn_b := Func(name="fn_b", interface=CallInterface(args=("b",)))),
        ir=file_ir[fn_b],
    )
    fn_c_target = IrTarget(
        symbol=(fn_c := Func(name="fn_c", interface=CallInterface(args=("c",)))),
        ir=file_ir[fn_c],
    )

    fn_a_actual = make_target_ir_call_tree(
        fn_a_target,
        environment=example_environment_f,
    )
    fn_a_expected = IrCallTreeNode(
        target=fn_a_target,
        edge_in=None,
        edges_out=[
            fn_a_to_fn_a_call := IrCall(
                caller=fn_a,
                symbol=Call(name="fn_a", args=CallArguments(args=("a",)), target=fn_a),
            ),
            fn_a_to_fn_b_call := IrCall(
                caller=fn_a,
                symbol=Call(name="fn_b", args=CallArguments(args=("a",)), target=fn_b),
            ),
        ],
        children=[
            IrCallTreeNode(
                target=fn_a_target,
                edge_in=fn_a_to_fn_a_call,
                edges_out=[
                    fn_a_to_fn_a_call,  # direct recursion
                    fn_a_to_fn_b_call,
                ],
                children=[],
            ),
            IrCallTreeNode(
                target=fn_b_target,
                edge_in=fn_a_to_fn_b_call,
                edges_out=[
                    # indirect recursion: a -> b -> a
                    fn_b_to_a_call := IrCall(
                        caller=fn_b,
                        symbol=Call(
                            name="fn_a",
                            args=CallArguments(args=("b",)),
                            target=fn_a,
                        ),
                    ),
                    fn_b_to_c_call := IrCall(
                        caller=fn_b,
                        symbol=Call(
                            name="fn_c",
                            args=CallArguments(args=("b",)),
                            target=fn_c,
                        ),
                    ),
                ],
                children=[
                    # from above indirect recursion
                    # thus this is a repeat of root with a different edge_in
                    # but without children (cycle elimination)
                    IrCallTreeNode(
                        target=fn_a_target,
                        edge_in=fn_b_to_a_call,
                        edges_out=[
                            fn_a_to_fn_a_call,
                            fn_a_to_fn_b_call,
                        ],
                        children=[],
                    ),
                    IrCallTreeNode(
                        target=fn_c_target,
                        edge_in=fn_b_to_c_call,
                        edges_out=[
                            fn_c_to_fn_b_call := IrCall(
                                caller=fn_c,
                                symbol=Call(
                                    name="fn_b",
                                    args=CallArguments(args=("c",)),
                                    target=fn_b,
                                ),
                            ),
                        ],
                        children=[
                            IrCallTreeNode(
                                target=fn_b_target,
                                edge_in=fn_c_to_fn_b_call,
                                edges_out=[fn_b_to_a_call, fn_b_to_c_call],
                                children=[],
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )
    assert fn_a_actual == fn_a_expected

    fn_b_actual = make_target_ir_call_tree(
        fn_b_target,
        environment=example_environment_f,
    )
    fn_b_expected = IrCallTreeNode(
        target=fn_b_target,
        edge_in=None,
        edges_out=[
            fn_b_to_a_call,
            fn_b_to_c_call,
        ],
        children=[
            IrCallTreeNode(
                target=fn_a_target,
                edge_in=fn_b_to_a_call,
                edges_out=[
                    fn_a_to_fn_a_call,
                    fn_a_to_fn_b_call,
                ],
                children=[
                    IrCallTreeNode(
                        target=fn_a_target,
                        edge_in=fn_a_to_fn_a_call,
                        edges_out=[
                            fn_a_to_fn_a_call,
                            fn_a_to_fn_b_call,
                        ],
                        children=[],
                    ),
                    IrCallTreeNode(
                        target=fn_b_target,
                        edge_in=fn_a_to_fn_b_call,
                        edges_out=[
                            fn_b_to_a_call,
                            fn_b_to_c_call,
                        ],
                        children=[],
                    ),
                ],
            ),
            IrCallTreeNode(
                target=fn_c_target,
                edge_in=fn_b_to_c_call,
                edges_out=[fn_c_to_fn_b_call],
                children=[
                    IrCallTreeNode(
                        target=fn_b_target,
                        edge_in=fn_c_to_fn_b_call,
                        edges_out=[
                            fn_b_to_a_call,
                            fn_b_to_c_call,
                        ],
                        children=[],
                    ),
                ],
            ),
        ],
    )
    assert fn_b_actual == fn_b_expected

    fn_c_actual = make_target_ir_call_tree(
        fn_c_target,
        environment=example_environment_f,
    )
    fn_c_expected = IrCallTreeNode(
        target=fn_c_target,
        edge_in=None,
        edges_out=[fn_c_to_fn_b_call],
        children=[
            IrCallTreeNode(
                target=fn_b_target,
                edge_in=fn_c_to_fn_b_call,
                edges_out=[
                    fn_b_to_a_call,
                    fn_b_to_c_call,
                ],
                children=[
                    IrCallTreeNode(
                        target=fn_a_target,
                        edge_in=fn_b_to_a_call,
                        edges_out=[
                            fn_a_to_fn_a_call,
                            fn_a_to_fn_b_call,
                        ],
                        children=[
                            IrCallTreeNode(
                                target=fn_a_target,
                                edge_in=fn_a_to_fn_a_call,
                                edges_out=[
                                    fn_a_to_fn_a_call,
                                    fn_a_to_fn_b_call,
                                ],
                                children=[],
                            ),
                            IrCallTreeNode(
                                target=fn_b_target,
                                edge_in=fn_a_to_fn_b_call,
                                edges_out=[
                                    fn_b_to_a_call,
                                    fn_b_to_c_call,
                                ],
                                children=[],
                            ),
                        ],
                    ),
                    IrCallTreeNode(
                        target=fn_c_target,
                        edge_in=fn_b_to_c_call,
                        edges_out=[fn_c_to_fn_b_call],
                        children=[],
                    ),
                ],
            ),
        ],
    )
    assert fn_c_actual == fn_c_expected

    _, stderr = capfd.readouterr()
    assert match_output(stderr, [])



from __future__ import annotations

# isort: off
from .function import FunctionResults
from .file import FileResults, FunctionName
from .cacheable import CacheableResults, make_cacheable_results

__all__ = [
    "FunctionResults",
    "FileResults",
    "FunctionName",
    "CacheableResults",
    "make_cacheable_results",
]



from __future__ import annotations

import ast
from importlib.util import find_spec
from pathlib import Path

import pytest

from rattr.models.symbol import (
    PYTHON_ATTR_ACCESS_BUILTINS,
    PYTHON_BUILTINS,
    AnyCallInterface,
    Builtin,
    Call,
    CallArguments,
    CallInterface,
    Class,
    Func,
    Import,
    Name,
    Symbol,
)
from rattr.module_locator.models import ModuleSpec
from rattr.module_locator.util import format_origin_for_os


class TestAttrsDerivedProperties:
    def test_eq_on_equivalent(self):
        assert Name("a") == Name("a")
        assert Name("a") == Name("a", "a")

    def test_eq_on_non_equivalent(self):
        assert Name("a") != Name("b")
        assert Name("a") != Name("a", "b")
        assert Name("a") != Import("a")

    def test_hash_on_identical(self):
        assert hash(Name("a")) == hash(Name("a"))

    def test_hash_on_equivalent(self):
        assert hash(Name("a")) == hash(Name("a", "a"))

    def test_hash_on_non_equivalent(self):
        assert hash(Name("a")) != hash(Name("b"))
        assert hash(Name("a")) != hash(Name("a", "b"))
        assert hash(Name("a")) != hash(Import("a"))

    @pytest.mark.parametrize(
        "symbol_fixture",
        [
            "simple_name",
            "simple_builtin",
            "simple_import",
            "simple_func",
            "simple_class",
            "simple_call",
        ],
    )
    def test_is_hashable(self, symbol_fixture: str, request: pytest.FixtureRequest):
        # It was observed that the derived hash may fail if an attribute type is
        # non-hashable
        symbol: Symbol = request.getfixturevalue(symbol_fixture)
        assert isinstance(hash(symbol), int)


class TestId:
    @pytest.mark.parametrize(
        "symbol_fixture",
        ["simple_name", "simple_builtin", "simple_func", "simple_class", "simple_call"],
    )
    def test_id(self, symbol_fixture: str, request: pytest.FixtureRequest):
        symbol: Symbol = request.getfixturevalue(symbol_fixture)
        assert symbol.id == symbol.name

    def test_id_for_direct_import(self):
        assert Import(name="pi", qualified_name="math.pi").id == "pi"

    def test_id_for_starred_import(self):
        assert Import(name="*", qualified_name="math").id == "math.*"


class TestIsCallable:
    @pytest.mark.parametrize(
        "symbol_fixture",
        ["simple_builtin", "simple_import", "simple_func", "simple_class"],
    )
    def test_is_callable(self, symbol_fixture: str, request: pytest.FixtureRequest):
        symbol: Symbol = request.getfixturevalue(symbol_fixture)
        assert symbol.is_callable

    @pytest.mark.parametrize("symbol_fixture", ["simple_name", "simple_call"])
    def test_is_not_callable(self, symbol_fixture: str, request: pytest.FixtureRequest):
        symbol: Symbol = request.getfixturevalue(symbol_fixture)
        assert not symbol.is_callable


class TestHasLocation:
    @pytest.mark.parametrize(
        "symbol_fixture",
        ["simple_name", "simple_func", "simple_class", "simple_import", "simple_call"],
    )
    def test_has_location(
        self,
        symbol_fixture: str,
        request: pytest.FixtureRequest,
        test_file: Path,
    ):
        symbol: Symbol = request.getfixturevalue(symbol_fixture)
        assert symbol.has_location
        assert symbol.location.defined_in == test_file

    @pytest.mark.parametrize(
        "symbol_fixture",
        ["simple_builtin"],
    )
    def test_does_not_have_location(
        self,
        symbol_fixture: str,
        request: pytest.FixtureRequest,
    ):
        symbol: Symbol = request.getfixturevalue(symbol_fixture)
        assert not symbol.has_location


class TestIsImport:
    @pytest.mark.parametrize("symbol_fixture", ["simple_import"])
    def test_is_import(self, symbol_fixture: str, request: pytest.FixtureRequest):
        symbol: Symbol = request.getfixturevalue(symbol_fixture)
        assert symbol.is_import

    @pytest.mark.parametrize(
        "symbol_fixture",
        ["simple_name", "simple_builtin", "simple_func", "simple_class", "simple_call"],
    )
    def test_is_not_import(self, symbol_fixture: str, request: pytest.FixtureRequest):
        symbol: Symbol = request.getfixturevalue(symbol_fixture)
        assert not symbol.is_import


class TestDerivedConstants:
    def test_python_builtins_is_populated(self):
        assert PYTHON_BUILTINS


class TestName:
    def test_default_basename(self):
        assert Name("my_name") == Name("my_name", "my_name")

    def test_derived_basename_is_base(self):
        name, basename = "*my.complex.attr_access[]", "my"
        assert Name(name).basename == "my"
        assert Name(name) == Name(name, basename)


class TestBuiltin:
    @pytest.mark.parametrize("name", PYTHON_ATTR_ACCESS_BUILTINS)
    def test_has_affect(self, name: str):
        assert Builtin(name).has_affect

    @pytest.mark.parametrize(
        "name", (b for b in PYTHON_BUILTINS if b not in PYTHON_ATTR_ACCESS_BUILTINS)
    )
    def test_does_not_have_affect(self, name: str):
        assert not Builtin(name).has_affect


class TestImport:
    def test_non_existant(self):
        b_func = Import("func", "b.func")

        assert b_func.name == "func"
        assert b_func.qualified_name == "b.func"

        assert b_func.module_name is None
        assert b_func.module_spec is None

    @pytest.mark.parametrize(
        "name,qualified_name,module_name",
        [
            ("math", "math", "math"),
            ("pi", "math.pi", "math"),
            ("join", "os.path.join", "os.path"),
        ],
    )
    def test_qualified_name(self, name, qualified_name, module_name):
        # NOTE
        # Use `find_spec` not `find_module_spec_fast` here as we are indirectly testing
        # the correctness of the latter's impl.
        stdlib_spec = find_spec(module_name)
        rattr_spec = (
            ModuleSpec(
                name=stdlib_spec.name,
                origin=format_origin_for_os(stdlib_spec.origin),
            )
            if stdlib_spec is not None
            else None
        )

        import_ = Import(name, qualified_name)

        assert import_.name == name
        assert import_.qualified_name == qualified_name

        assert import_.module_name == module_name
        assert import_.module_spec == rattr_spec


class TestFunc:
    def test_name_excludes_call_brackets(self):
        fn = "my_lib.func"

        without_call_brackets = Func(fn, interface=AnyCallInterface())
        with_call_brackets = Func(f"{fn}()", interface=AnyCallInterface())

        assert without_call_brackets == with_call_brackets
        assert without_call_brackets.name == with_call_brackets.name == fn

    def test_from_fn_def(self):
        fn = ast.parse("def fn(a, /, b=None, *, c=False): pass").body[0]
        async_fn = ast.parse("async def fn(a, /, b=None, *, c=False): pass").body[0]

        fn_spec = {
            "name": "fn",
            "interface": CallInterface(
                posonlyargs=["a"],
                args=["b"],
                kwonlyargs=["c"],
            ),
        }
        expected = Func(**fn_spec, token=fn)
        expected_async = Func(**fn_spec, token=async_fn, is_async=True)

        assert Func.from_fn_def(fn) == expected
        assert Func.from_fn_def(async_fn) == expected_async

    def test_from_fn_def_with_vararg_and_kwarg(self):
        fn = ast.parse("def fn(a, /, b=None, *c, d=False, e=1, **f): pass").body[0]

        fn_spec = {
            "name": "fn",
            "interface": CallInterface(
                posonlyargs=["a"],
                args=["b"],
                vararg="c",
                kwonlyargs=["d", "e"],
                kwarg="f",
            ),
            "token": fn,
        }
        expected = Func(**fn_spec)

        assert Func.from_fn_def(fn) == expected


class TestClass:
    def test_name_excludes_call_brackets(self):
        cls = "my_lib.MyClass"

        without_call_brackets = Class(cls, interface=AnyCallInterface())
        with_call_brackets = Class(f"{cls}()", interface=AnyCallInterface())

        assert without_call_brackets == with_call_brackets
        assert without_call_brackets.name == with_call_brackets.name == cls

    def test_with_init(self):
        cls = Class("MyClass")
        init = ast.parse("def __init__(self, a, *, b=None): pass").body[0]

        expected = Class(
            "MyClass",
            interface=CallInterface(
                posonlyargs=[],
                args=["self", "a"],
                kwonlyargs=["b"],
            ),
        )

        assert cls.with_init(init) == expected


class TestCall:
    def test_name_excludes_call_brackets(self):
        fn = "my_lib.func"

        without_call_brackets = Call(fn, args=CallArguments())
        with_call_brackets = Call(f"{fn}()", args=CallArguments())

        assert without_call_brackets == with_call_brackets
        assert without_call_brackets.name == with_call_brackets.name == fn

    def test_from_call(self, constant):
        call = ast.parse("fn(a, b, c=var, d='yes')").body[0].value

        expected = Call(
            name="fn",
            args=CallArguments(
                args=["a", "b"],
                kwargs={"c": "var", "d": constant},
            ),
            target=None,
            token=call,
        )

        assert Call.from_call("fn", call, target=None) == expected
